# =========================================================================
#
# start ADAPTOR (if needed) and connect to it - interface routine
#

proc start_adaptor {} {
	global adaptorinfo sidinfo maininfo
	match_max -d 10000
	
	if { [info exists adaptorinfo(program)] && 
	     (![string compare $adaptorinfo(program) ""] == 0) } {
		## open a pty from which we will read the stderr output
		## of ADAPTOR
		## (we will redirect the stderr of ADAPTOR to this pty)
		#if {[catch {spawn -pty} foo]} {
		#	fatal "spawn for stderr of $adaptorinfo(program) fails: $foo"
		#}
		#set sidinfo(adaptor,stderr) $spawn_id
		#set adaptorinfo(pty,stderr) $spawn_out(slave,name)
		#send_error "DEBUG: ADAPTOR Stderr match_max=[match_max]\n"
		#set sidinfo(sid,name,$sidinfo(adaptor,stderr)) ADAPTOR
		#lappend sidinfo(bg) $sidinfo(adaptor,stderr)
		
		# build the command to start the ADAPTOR:
		# - we disable echo-ing through the shell stty command 
		# - the adaptor runs in the 'hexcontext' context
		# - the stderr of the adaptor is redirected to the pty
		#   we spawned above
		if {[info exists adaptorinfo(program,run,dir)]} {
			set cdcmd "cd $adaptorinfo(program,run,dir)"
			set program $adaptorinfo(program)
		} elseif {[info exists adaptorinfo(program,run,dir,default)]} {
			set cdcmd "cd $adaptorinfo(program,run,dir,default)"
			set program $adaptorinfo(program)
		} else {
			set cdcmd "cd [file dirname $adaptorinfo(program)]"
			set program [file join . [file tail $adaptorinfo(program)]]
		}
		set arglist [concat $adaptorinfo(flags) $maininfo(configfile)]
		set args [join $arglist " "]
		# set command [list spawn -noecho "/bin/sh" -c "stty -echo; exec 2> $adaptorinfo(pty,stderr); $cdcmd; exec $adaptorinfo(context) $program $args"] 
		set command [list spawn -noecho "/bin/sh" -c "stty -echo; $cdcmd && exec torx-stdiomerge \"\" STDERR: exec $adaptorinfo(context) $program $args"] 
		# set command [list spawn -noecho "/bin/sh" -c "torx-stdiomerge \"\" STDERR: $cdcmd \\; exec $adaptorinfo(context) $program $args"] 
		debugLvl 0 "$command\n"
		catch { send_error "DEBUG: Starting adaptor: ($command)\n" }
		if {[catch $command foo]} {
			fatal "spawn for $adaptorinfo(program) fails: $foo"
		}
		set sidinfo(adaptor) $spawn_id
		set adaptorinfo(pid) [exp_pid -i $sidinfo(adaptor)]
		send_error "DEBUG: ADAPTOR match_max=[match_max]\n"
		io_add_peer adaptor $sidinfo(adaptor) $adaptorinfo(timeout)
		io_config $maininfo(timeout)

#		expect_background {
#			-i $sidinfo(adaptor,stderr) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
#				set txt $expect_out(1,string)
#				send_error "ADAPTOR Debug: $txt\n"
#			}
#			-i $sidinfo(adaptor,stderr) -re "(\[^\r\n]*)\r*\n" {
#				set l $expect_out(1,string)
#				send_error "ADAPTOR Stderr: $l\n"
#			}
#		}

		expect_before {
			-i $sidinfo(adaptor) full_buffer {
				warning "ADAPTOR: Buffer Overflow!!!\n"
				set txt $expect_out(buffer)
				warning "ADAPTOR: Buffer Contents Skipped: $txt\n"
				fatal "ADAPTOR: Buffer Contents Skipped: $txt\n"
				exit
 			}
			-i $sidinfo(adaptor) eof {
				global sidinfo adaptorinfo
				if {[info exists sidinfo(adaptor)]} {
					unset sidinfo(adaptor)
	 			}
				catch {unset adaptorinfo(pid)}
				io_rm_peer adaptor
				if {![cleaning_up]} {
					fatal "End-of-file from adaptor\n"
		 		}
		 	}
			-i $sidinfo(adaptor) -re "^(ERROR (\[^\r\n]*)\r*\n)+" {
				foreach l [split $expect_out(buffer) "\n"] {
					if {[regexp -- "^ERROR (\[^\r\n\]*)" $l match sub]} {
						send_error "ADAPTOR Error: $sub\n"
					} else {
						send_error "ADAPTOR Error: $l\n"
					}
				}
			}
		}
		expect_after {
			-i $sidinfo(adaptor) -re "^C_\[^\r\n]*\r*\n" {
				# skip. This is echo'ed back from the adaptor module
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^A_QUIT\[ \t]*\r*\n" {
				set result [list error]
			}
			-i $sidinfo(adaptor) -re "^A_WARNING(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				warning "ADAPTOR: $txt\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^A_MSG(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				message "ADAPTOR: $txt\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^A_NOTIMPLEMENTED(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				nonfatal "ADAPTOR unimplemented command: $txt\n"
				set result [list error ]
			}
			-i $sidinfo(adaptor) -re "^A_ERROR\[ \t]\[\t ]*UnknownCommand:\[ \t]\[\t ]*(\[^\r\n]\[^\r\n]*)\[ \t]*\r*\n" {
				# This regexp tries to catch the 'not-implemented'
				# answers from old adaptor components (i.e. those that do not
				# yet implement the A_NOTIMPLEMENTED answer)
				# to allow us to continue working, instead of interpreting
				# it as a fatal error.
				# WARNING: this regexp 'knows' that old adaptor components
				# use this particular format in case of unknown commands
				# (this works at least for the LOTOS and Promela components):
				#   A_ERROR UnknownCommand: command-that-was-not-implemented
				# (including the :)
				# This just might fail for other adaptor componts that
				# use a slightly different error message...
				set txt $expect_out(1,string)
				warning "command not implemented in ADAPTOR: $txt\n"
				set result [list error ]
			}
			-i $sidinfo(adaptor) -re "^A_ERROR(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				fatal "Error ADAPTOR: $txt\n"
			}
			-i $sidinfo(adaptor) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				send_error "ADAPTOR Debug: $txt\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^A_LOG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				send_error "ADAPTOR Log: $txt\n"
				utlog_tofile_string ADAPTOR $txt
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^(A_\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				warning "Unexpected ADAPTOR Answer Ignored : $txt\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^STDERR:A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
				set txt $expect_out(1,string)
				send_error "ADAPTOR Debug: $txt\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^STDERR:(\[^\r\n]*)\r*\n" {
				set l $expect_out(1,string)
				send_error "ADAPTOR Stderr: $l\n"
				exp_continue
			}
			-i $sidinfo(adaptor) -re "^(\[^\r\n]*)\r*\n" {
				set l $expect_out(buffer)
				if {[regexp -- "^(\[^\r\n\]*)" $l match sub]} {
					send_error "ADAPTOR: $sub\n"
				} else {
					send_error "ADAPTOR: $l\n"
				}
				exp_continue
			}
		}

	} else {
		set adaptorinfo(pid) ""
	}

}

# =========================================================================
