# =========================================================================
#
# main loop
#

proc main_loop {} {
	global maininfo
	set step_failure 0
	log_time_and_state
	while {[main_go_on]} {
		if {$step_failure} {
			set step_failure 0
		} else {
			incr maininfo(nrofstepsdone)
			set tr_event [uttrace_read_event]
		}
		set cmd_interp_event [utcmdline_interpret $tr_event]
		set kind [lindex $cmd_interp_event 0]
		set chan [lindex $cmd_interp_event 1]
		set event [lindex $cmd_interp_event 2]
		if {[regexp {^([^[]*)[ \t]+\[([^]]*)\]}  $event match ev preds]} {
			debugLvl 2 "split ($event) -> ($ev) ($preds)"
			set cmd_interp_event [list $kind $chan $ev]
			set cmd_interp_preds $preds
		} else {
			set cmd_interp_preds ""
		}
		# stop tracing if the user decided not to follow the trace
		# but we don't have to do that here
		# uttrace_check_event [uttrace_event $tr_event] [uttrace_event $cmd_interp_event]
		if {$maininfo(stopnow)} break ;# in case the command line demands it
		if {[llength [uttrace_iokind $tr_event]] > 0} {
			# send_user "Trace: [uttrace_iokind $tr_event] [uttrace_event $tr_event]\n"
		}
		set tr_iokind_chan [uttrace_iokind $cmd_interp_event]
		set ad_iokind_chan [adaptor_get_iokind $tr_iokind_chan]
		set iokind_chan [spec_get_iokind $ad_iokind_chan]
		set chan [lindex $iokind_chan 1]
		switch -- [lindex $iokind_chan 0] \
		  stop {
		  	# undo increment we just did. nasty, but necessary
		  	incr maininfo(nrofstepsdone) -1
		  	set maininfo(endofspec) 1
		  	utlog_tofile_eospec
			catch {send_error "End of SPEC (sink state)\n"}
		} input {
			uttrace_check_iokind $iokind_chan [uttrace_iokind $tr_event]
			set event_to_try [uttrace_event [spec_choose_from_menu $cmd_interp_event $cmd_interp_preds]]
			set input [spec_get_input $chan $event_to_try $cmd_interp_preds]
			debugLvl 2 "event_to_try=$event_to_try"
			debugLvl 2 "cmd_interp_preds=$cmd_interp_preds"
			debugLvl 2 "input=$input"
			switch -- [lindex $input 0] \
			  input_ok {
				set chan_spec [lindex $input 1]
				set ev_spec [lindex $input 2]
				set susp_spec [lindex $input 3]
				set preds_spec [lindex $input 4]
				# BUG: actually, to see if we have to instantiate,
				# we should not be looking at the predicates,
				# but at the free variables
				# (there may be variables without constraints)
				# (same comment applies below)
				# However, currently we do not pass the freevars over the interfacce
				if {[string compare $preds_spec ""] != 0 } {
					set inst_result [instantiate $preds_spec]
					switch -- [lindex $inst_result 0] \
					  inst_ok {
						set inst_preds_spec [lindex $inst_result 1]
					} inst_error {
						set step_failure 1
						set inst_preds_spec [lindex $inst_result 1]
						set reason [lindex $inst_result 2]
						set expl [lindex $inst_result 3]
						send_user "instantiator: unsuccesful instantiation: $reason: $ev_spec \[ $preds_spec \]\n"
						send_user "instantiator: explanation: $expl\n"
						send_user "instantiator: (start of unsuccesful instantiation)\n"
						send_user "instantiator: user action: $event_to_try\n"
						send_user "instantiator: user predicates: $cmd_interp_preds\n"
						send_user "instantiator: input action channel: $chan_spec\n"
						send_user "instantiator: input action event: $ev_spec\n"
						send_user "instantiator: input action predicates: $preds_spec\n"
						send_user "instantiator: predicates instantiation: $inst_preds_spec\n"
						send_user "instantiator: (end of unsuccesful instantiation)\n"
					}
				} else {
					set inst_preds_spec $preds_spec
				}
				debugLvl 2 "chan_spec=$chan_spec"
				debugLvl 2 "ev_spec=$ev_spec"
				debugLvl 2 "preds_spec=$preds_spec"
				debugLvl 2 "inst_preds_spec=$inst_preds_spec"
				if {! $step_failure &&
				    [string compare $preds_spec ""] != 0 &&
				    [string compare $preds_spec $inst_preds_spec] != 0} {
					set checked_input [spec_get_input $chan_spec $ev_spec $inst_preds_spec]
					switch -- [lindex $checked_input 0] \
					  input_ok {
						set chan_spec [lindex $checked_input 1]
						set ev_spec [lindex $checked_input 2]
						set susp_spec [lindex $checked_input 3]
						set preds_spec [lindex $checked_input 4]
						debugLvl 2 "chan_spec=$chan_spec"
						debugLvl 2 "ev_spec=$ev_spec"
						debugLvl 2 "preds_spec=$preds_spec"
					} input_error {
					  	set step_failure 1
						if {[llength $input] > 1} {
							set reason [join [lrange $input 1 end] " "]
							send_user "Primer: instantiation yields invalid input action: $reason\n"
						} else {
							send_user "Primer: instantiation yields invalid input action\n"
						}
						send_user "primer: result of instantiation: $checked_input\n"
						send_user "primer: (start of invalid action)\n"
						send_user "primer: input action channel: $chan_spec\n"
						send_user "primer: input action event: $ev_spec\n"
						send_user "primer: input action predicates: $preds_spec\n"
						send_user "primer: input action instantiation: $inst_preds_spec\n"
						send_user "primer: (end of invalid action)\n"
					} error {
						fatal "error from specification\n"
					}
				}
				if {! $step_failure} {
					utlog_tostore "input" $chan $ev_spec $susp_spec $preds_spec
					set input_sent [adaptor_put_input $chan $ev_spec $preds_spec]
					switch -- [lindex $input_sent 0] \
					  input_ok {
						set chan_sent [lindex $input_sent 1]
						set ev [lindex $input_sent 2]
						set susp [lindex $input_sent 3]
						set preds [lindex $input_sent 4]
						set pco [lindex $input_sent 5]
						set tmstmp [lindex $input_sent 6]
						set conc [lindex $input_sent 7]
						utlog_tofile "input" $chan $ev $susp $preds $pco $tmstmp $conc
						utlog_touser "input" $chan $ev $susp $preds $pco ""
						check_input $chan $ev $susp $preds
						# NOTE: in the future, uttrace_check_event should no
						# longer be necessary, because if the trace-event is not
						# possible, we will be in case 'input_error'.
						uttrace_check_event $ev [uttrace_event $tr_event]
					} input_error {
						set reason [lindex $input_sent 1]
						set predicates [lindex $input_sent 2]
						set explanation [lindex $input_sent 3]
						if {[llength $input_sent] > 1} {
							send_user "Adaptor: input action unsuccessful: $reason \[$predicates\] ($explanation)\n"
						} else {
							send_user "Adaptor: input action unsuccessful\n"
						}
						if {[string compare $preds_spec ""] == 0} {
							set preds $predicates
						} else {
							set preds "$preds_spec $predicates"
						}
						utlog_tofile_refusal "input" $chan $ev_spec 1 $preds "" "" "" $reason $explanation
						utlog_touser "input" $chan $ev_spec 1 $preds "" $reason
						send_user "adaptor: (start of unsuccesful input action)\n"
						send_user "adaptor: input action event: $ev_spec\n"
						send_user "adaptor: input action preds: $preds_spec\n"
						send_user "adaptor: input action channel: $chan\n"
						send_user "adaptor: reason: $reason\n"
						send_user "adaptor: preds: $predicates\n"
						send_user "adaptor: explanation: $explanation\n"
						send_user "adaptor: (end of unsuccesful input action)\n"
					} error {
						fatal "error from adaptor\n"
					}
				}
			} input_error {
			  	set step_failure 1
				if {[llength $input] > 1} {
					set reason [join [lrange $input 1 end] " "]
					send_user "Primer: input action unsuccessful: $reason\n"
				} else {
					send_user "Primer: input action unsuccessful\n"
				}
				send_user "primer: (start of unsuccesful input action)\n"
				send_user "primer: input action event: $event_to_try\n"
				send_user "primer: input action channel: $chan\n"
				send_user "primer: (end of unsuccesful input action)\n"
				send_user "DEBUG cmd_interp_event=($cmd_interp_event) tr_iokind_chan=($tr_iokind_chan) ad_iokind_chan=($ad_iokind_chan) iokind_chan=($iokind_chan)\n"
			} error {
				fatal "error from specification\n"
			}
		} output {
			uttrace_check_iokind $iokind_chan [uttrace_iokind $tr_event]
			utlog_tostore "output" $chan "" "" ""
			set output [adaptor_get_output $chan]
			switch -- [lindex $output 0] \
			  output_ok {
				set chan_rcvd [lindex $output 1]
				set ev [lindex $output 2]
				set susp [lindex $output 3]
				set preds [lindex $output 4]
				set pco [lindex $output 5]
				set tmstmp [lindex $output 6]
				set conc [lindex $output 7]
				utlog_tofile "output" $chan $ev $susp $preds $pco $tmstmp $conc
				# I'm not sure whether we should change the label (ev) here at all
				# I'm tempted to rip it out, but do not know what harm it might do.
				# To be on the safe side, we continue to change the ev
				# but (to make timed testing work again) only when
				# the adapter did not provide an event
				if {$susp && [string compare $ev ""] == 0} {
					if {[info exists maininfo(parsed-label,delta)]} {
						set ev $maininfo(parsed-label,delta)
					}
				}
				utlog_touser "output" $chan $ev $susp $preds $pco ""
				uttrace_check_event $ev [uttrace_event $tr_event]
				check_output $chan $ev $susp $preds
			} output_error {
				if {[llength $output] > 1} {
					set reason [join [lrange $input 1 end] " "]
					send_user "Adaptor: output action unsuccessful: $reason\n"
				} else {
					send_user "Adaptor: output action unsuccessful\n"
				}
			} error {
				fatal "error from adaptor\n"
			}
		} error {
			fatal "error from specification\n"
		} default {
		}
		# also done inside check_{input,output} in case of {input,output}_error
		if {(! $step_failure) && (!$maininfo(endofspec))} {
			log_time_and_state
		}
	}
	catch {send_error "End of Test\n"}
}

# =========================================================================
