proc explorer_init {sim} {
	global exp

	set exp(sim) $sim
	set exp(nr) 0
	set exp(cur,s) ""
	set exp(init) [get_state] ;# side effect: initializes exp(cur,s)

	return ""
}

proc explorer_main {sim} {
	global user_spawn_id
	global exp
	global io
	global timeout

	while 1 {
		set timeout $io(timeout,$sim)
		expect -i $user_spawn_id -re "(\[^\n\r]*)\r*\n" {
			set l $expect_out(1,string)
			explorer_handle_line $l
		}
	}
}

# we map between the event-numbers that we use
# in the torx-explorer interface (and that we invent
# ourself, using exp(nr)) and the 'real' state numbers
# (that we get from the primer we provide the interface for)
# the mapping is implemented by assoc array exp(nr2state,$snr)
# event numbers are 'n' or 'snr'
# state numbers are 's' or 'd'
# assoc array exp(tmap,$s,$l,$d) assumes that the primer
# is a real LTS: there is at most one <s,l,d> in the trans. relation.

proc explorer_start {} {
	global exp
	set n $exp(nr)
	set d $exp(init)
	set exp(nr2state,$n) $d
	if {![info exists exp(identical,$d)]} {
		set exp(identical,$d) $n
	}
	incr exp(nr)
	return [unparse_state $n]
}

proc explorer_unfold {smsg} {
	global exp
	if {![regexp {\.([0-9]+)$} $smsg m snr]} {
		set snr $smsg
	}
	if {[string compare $snr ""] == 0} {
		return -code error "e needs argument (usage: e srcid)"
	}
	if {![info exists exp(nr2state,$snr)]} {
		return -code error "non-existant nr: $snr"
	}
	set s $exp(nr2state,$snr)
	set evs {}
	set in [get_inputs $s]
	set out [get_outputs $s]
	foreach ev [concat $in $out] {
		set l $ev
		set d [get_succ $s $ev]
		if {[info exists exp(tmap,$s,$l,$d)]} {
			set n $exp(tmap,$s,$l,$d)
			puts stderr "found: $snr ($s) - ($l) -> $d : $n"
		} else {
			set n $exp(nr)
			incr exp(nr)
			set exp(nr2state,$n) $d
			set exp(tmap,$s,$l,$d) $n
			puts stderr "set: $snr ($s) - ($l) -> $d : $n"
		}
		if {![info exists exp(identical,$d)]} {
			set exp(identical,$d) $n
			set ew ""
			puts stderr "set: $d = $exp(identical,$d)"
		} elseif {[string compare $exp(identical,$d) $n] != 0} {
			set ew S[set d].$exp(identical,$d)
			puts stderr "found: $d = $exp(identical,$d)"
		} else {
			set ew ""
			puts stderr "found: $d = $exp(identical,$d)"
		}
		lappend evs [list S[set d].[set n] $l $ew]
	}
	if {[llength $evs] <= 0} {
		warningMsg "empty menu"
		# return -code error "empty menu"
	}
	set res {}
	foreach e $evs {
		lappend res [unparse_event $e]
	}
	return $res
}

proc explorer_handle_line {line} {
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [parse_line $line]
		debugMsg "Parsed: ($parsed_line)"
		set cmd_keyword [lindex $parsed_line 0]
		switch -- $cmd_keyword {
		    q {
			set cont 0
			puts "Q"
		  } r {
			set initial [explorer_start]
			puts "R [join $initial \t]"
		  } e {
			set smsg [lindex $parsed_line 1]
			
			if {[catch {explorer_unfold $smsg} evs]} {
				puts "E0 $evs"
			} else {
				puts "EB"
				foreach e $evs {
					puts "Ee\t[join $e \t]"
				}
				puts "EE"
			}
		  } d {
			# nothing to do here; in  general: delete list of events/states
			puts "D"
		  } default {
			puts "A_ERROR UnknownCommand: $cmd_keyword"
		  }
		}
	}
}

# =========================================================================

proc parse_line {line} {
	if {[regexp {^(.)([ \t]+(.*))?$} $line match cmd dummy remains]} {
		return [list $cmd $remains]
	}
}

# =========================================================================

proc unparse_event {e} {
	set e_repr [lindex $e 0]
	set preds_solved 1 
	set label [lindex $e 1]
	set npreds ""
	set nfreevars ""
	set elsewhere [lindex $e 2]
	set visible [isvisible $label]
	return [list $e_repr $visible $preds_solved $label $npreds $nfreevars $elsewhere]
}

proc unparse_state {e} {
	global exp
	set d $exp(nr2state,$e)
	set state S$d.$e
	set preds_solved 1
	set preds ""
	set freevars ""
	set npreds ""
	set nfreevars ""
	set elsewhere ""
	return [list $state $preds_solved $npreds $nfreevars $elsewhere]
}

proc isvisible {label} {
	return 1
}

# =========================================================================
