# we explictly change primer state in goto_state,
# we implictly chage primer state in get_succ,
# we get the primer state using get_state.
# get_state and goto_state share a common variable
# to avoid unnessary state change requests to the primer.

proc get_state {} {
	global exp
	io_send_command $exp(sim) STATEID
	io_receive_answer $exp(sim) {
		-singleline {
			STATEID {super init trans} {
				set exp(cur,s) $super
				puts stderr "STATEID super $super init $init trans $trans"
				return $super
			}
			{} {} {
				return -code error "get_state: unexpected : $line"
			}
		}
	}
}

proc goto_state {s} {
	global exp
	# debugMsg "goto_state $s"
	if {[string compare $exp(cur,s) $s] == 0} {
		return
	}
	io_send_command $exp(sim) GOTO [list \
			[list super $s]          \
		]
	io_receive_answer $exp(sim) {
		-singleline {
			GOTO_OK {} {
				set exp(cur,s) $s
			}
			GOTO_ERROR {} {
				return -code error "goto_state: $line"
			}
		}
	}
}

proc get_inputs {s} {
	global exp
	set events {}
	goto_state $s
	io_send_command $exp(sim) INPUTS [list \
			[list channel in]          \
		]
	io_receive_answer $exp(sim) {
		-multiline INPUTS {
			EVENT {event} {
				set exp(kind,$event) INPUT
				lappend events $event
			}
		}
	}
	return $events
}

proc get_outputs {s} {
	global exp
	set events {}
	goto_state $s
	io_send_command $exp(sim) OUTPUTS [list \
			[list channel out]          \
		]
	io_receive_answer $exp(sim) {
		-multiline OUTPUTS {
			EVENT {event} {
				set exp(kind,$event) OUTPUT
				lappend events $event
			}
		}
	}
	return $events
}

proc get_succ {s ev} {
	global exp
	if {![info exists exp(trans,$s,$ev)]} {
		goto_state $s
		io_send_command $exp(sim) $exp(kind,$ev) [list \
				[list event $ev]          \
			]
		io_receive_answer $exp(sim) {
			-singleline  {
				$exp(kind,$ev)_OK {event} {
					set exp(trans,$s,$ev) [get_state]
					puts stderr "get_succ: $s \"$ev\" : $exp(trans,$s,$ev)"
					# debugMsg "get_succ: $s \"$ev\" : $exp(trans,$s,$ev)"
				}
				$exp(kind,$ev)_ERROR {} {
					return -code error "get_succ: $s \"$ev\" : $exp(kind,$ev)_ERROR $line"
				}
			}
		}
	}
	return $exp(trans,$s,$ev)
}

