# =========================================================================
# $Id: main_loop.tcl,v 1.1.2.6 2000/10/06 12:18:01 belinfan Exp $
# =========================================================================

proc explorer_interface {} {
	if {[catch {do_explorer_interface} foo]} {
		bgerror $foo
	}
}
proc do_explorer_interface {} {
	explorer_init
#	disable_selection_exportation
#	explorer_start ;# compute initial menu
#	restore_selection_exportation
	explorer_main
}

proc explorer_main {} {
	## Currently, we prefere to use blocking IO
	## fconfigure -nonblocking stdin
	fileevent stdin readable explorer_try_handle_line
}

# ====================================

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc bgerror {msg} {
	global errorInfo
	debugMsg "Fatal InternalError: $msg"
	set lines [split $errorInfo \n]
	debugMsg "Stack Begin"
	foreach l $lines {
		debugMsg "Stack: $l"
	}
	debugMsg "Stack End"
	# exit
}


# ====================================

proc explorer_try_handle_line {} {
	fileevent stdin readable ""
	set stop 0
	set skip 0
	set line ""
	if {[gets stdin line] < 0} {
		if {[fblocked stdin]} {
			set skip 1
		} elseif {[eof stdin]} {
			set line "q"
			set stop 1
		}
	} 
	if {! $skip} {
		disable_selection_exportation
		set speed [time {explorer_handle_line $line}]
		# puts "A_DEBUG time: $speed"
		restore_selection_exportation
	}
	if {! $stop} {
		fileevent stdin readable explorer_try_handle_line
	}
}

# ====================================

proc explorer_handle_line {line} {
	# debugMsg "Line: ($line)"
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [parse_line $line]
		debugMsg "Parsed: ($parsed_line)"
		set cmd_keyword [lindex $parsed_line 0]
		switch -- $cmd_keyword {
		    q {
		    	set cont 0
			   	puts "Q"
		  } r {
		  		set initial [explorer_start]
		  		puts "R [join $initial \t]"
		  } e {
		  		set s [lindex $parsed_line 1]
		  		
		  		if {[catch {explorer_unfold $s} evs]} {
			  		puts "E0 $evs"
		  		} else {
			  		puts "EB"
			  		foreach e $evs {
			  			puts "Ee\t[join $e \t]"
			  		}
			  		puts "EE"
			  	}
		  } p {
		  		set preds [lindex $parsed_line 1]
		  		set res [join [explorer_solve $preds] "\t"]
		  		puts "P $res"
		  } m {
		  		set ev_evs [split [lindex $parsed_line 1]]
		  		set ev [lindex $ev_evs 0]
		  		set evs [lrange $ev_evs 1 end]
		  		if {![catch {explorer_match_events $ev $evs} res]} {
		  			puts "M [join $res \t]"
		  		} else {
		  			puts "M0 $res"
		  		}
		  } d {
		  		# nothing to do here; in  general: delete list of events/states
		  		puts "D"
		  } i {
		  		set ev_preds [split_in_two_on_ws [lindex $parsed_line 1]]
		  		set ev [lindex $ev_preds 0]
		  		set preds [lindex $ev_preds 1]
				# debugMsg "instantiate: ($ev)($preds)"
				set res [explorer_instantiate $ev $preds]
				if {[lindex $res 0]} {
					puts "I\t[join [lindex $res 1] \t]"
				} else {
		  			puts "I0\t[join [lindex $res 1] \t]"
				}
		  } default {
				puts "A_ERROR UnknownCommand: $cmd_keyword"
		  }
		}
	}
}

# =========================================================================

proc parse_line {line} {
	if {[regexp {^(.)([ \t]+(.*))?$} $line match cmd dummy remains]} {
		return [list $cmd $remains]
	}
}

proc split_in_two_on_ws {line} {
	if {[regexp {^([^ \t]+)([ \t]+(.*))?$} $line match cmd dummy remains]} {
		return [list $cmd $remains]
	}
}

# =========================================================================
