# =========================================================================
# $Id: autexp.tcl,v 1.1.2.12 2007/07/05 16:01:10 belinfan Exp $
# =========================================================================

# ====================================
# Explorer Tcl API Interface routines

proc explorer_init {f} {
	global aut
	if {[catch {open $f r} fid]} {
		return "failed to open $f: $fid"
	}
	set nr 0
	if {[gets $fid line] >= 0} {
		incr nr
		set line [string trim $line]
		if {[regexp {^des[ \t]*\([ \t]*([^ \t,]+)[ \t]*,[ \t]*([^ \t,]+)[ \t]*,[ \t]*([^ \t\)]+)[ \t]*\)$} $line all init trans states]} {
			# puts stderr "$init $trans $states"
			set aut(init) $init
			if {[string compare $aut(init) 0] != 0} {
				warningMsg "non-standard .aut (some programs may expect initial state to be '0'): $init"
			} elseif {![regexp {^[0-9]+$} $aut(init)]} {
				warningMsg "non-standard .aut input (some programs may expect states to be numbers): $init"
			}
		} elseif {[string compare $line ""] == 0} {
		} else {
			return "failed to parse aut header:$nr: $line"
		}
	} else {
		return "failed to read $f: $fid"
	}
	set enr 0
	while {[gets $fid line] >= 0} {
		incr nr
		set line [string trim $line]
		if {[regexp {^\([ \t]*([^ \t,]+)[ \t]*,[ \t]*"([^"]*)"[ \t]*,[ \t]*([^ \t\)]+)[ \t]*\)$} $line all src label dst] ||
		    [regexp {^\([ \t]*([^ \t,]+)[ \t]*,[ \t]*'([^']*)'[ \t]*,[ \t]*([^ \t\)]+)[ \t]*\)$} $line all src label dst] ||
		    [regexp {^\([ \t]*([^ \t,]+)[ \t]*,[ \t]*([^,]*)[ \t]*,[ \t]*([^ \t\)]+)[ \t]*\)$} $line all src label dst]} {
		    lappend aut(trans,$src) [list [list e$enr $label] $dst]
		    lappend aut(trans) [list $src  e$enr $label $dst]
		    if {![info exists aut(trans,$dst)]} {
			set aut(trans,$dst) {}
		    }
		    incr enr
		} elseif {[string compare $line ""] == 0} {
		} else {
			return "failed to parse aut line:$nr: $line"
		}
	}
	if {![info exists aut(init)]} {
		return "failed to read aut file (no header)"
	}
	if {![info exists aut(trans)]} {
		return "failed to read aut file (no transitions)"
	}
	set aut(nr) 0
	return ""
}

proc explorer_start {} {
	global aut
	set n $aut(nr)
	set d $aut(init)
	set aut(nr2state,$n) $d
	if {![info exists aut(identical,$d)]} {
		set aut(identical,$d) $n
	}
	incr aut(nr)
	return [unparse_state $n enil]
}

proc explorer_unfold {s} {
	global aut
	if {![regexp {\.([0-9]+)$} $s m src]} {
		set src $s
	}
	if {[string compare $src ""] == 0} {
		return -code error "e needs argument (usage: e srcid)"
	}
	if {![info exists aut(nr2state,$src)]} {
		return -code error "non-existant nr: $src"
	}
	set srcst $aut(nr2state,$src)
	set evs {}
	foreach t $aut(trans,$srcst) {
		set e [lindex [lindex $t 0] 0]
		set l [lindex [lindex $t 0] 1]
		set d [lindex $t 1]
		if {[info exists aut(tmap,$srcst,$e,$l,$d)]} {
			set n $aut(tmap,$srcst,$e,$l,$d)
			# puts stderr "found: $src ($srcst) - $e ($l) -> $d : $n"
		} else {
			set n $aut(nr)
			incr aut(nr)
			set aut(nr2state,$n) $d
			set aut(tmap,$srcst,$e,$l,$d) $n
			# puts stderr "set: $src ($srcst) - $e ($l) -> $d : $n"
		}
		if {![info exists aut(identical,$d)]} {
			set aut(identical,$d) $n
			set ew ""
			# puts stderr "set: $d = $aut(identical,$d)"
		} elseif {[string compare $aut(identical,$d) $n] != 0} {
			set ew S[set d]_[set e].$aut(identical,$d)
			# puts stderr "found: $d = $aut(identical,$d)"
		} else {
			set ew ""
			# puts stderr "found: $d = $aut(identical,$d)"
		}
		lappend evs [list S[set d]_[set e].[set n] $l $ew]
	}
	if {[llength $evs] <= 0} {
		warningMsg "empty menu"
		# return -code error "empty menu"
	}
	set res {}
	foreach e $evs {
		lappend res [unparse_event $e]
	}
	return $res
}

# =========================================================================
# Explorer Tcl API Utility routines

proc unparse_event {e} {
	set e_repr [lindex $e 0]
	set preds_solved 1 
	set label [lindex $e 1]
	set npreds ""
	set nfreevars ""
	set elsewhere [lindex $e 2]
	set visible [isvisible $label]
	return [list $e_repr $visible $preds_solved $label $npreds $nfreevars $elsewhere]
}

proc unparse_state {e evname} {
	global aut
	set d $aut(nr2state,$e)
	set state S[set d]_[set evname].$e
	set preds_solved 1
	set preds ""
	set freevars ""
	set npreds ""
	set nfreevars ""
	set elsewhere ""
	return [list $state $preds_solved $npreds $nfreevars $elsewhere]
}

proc isvisible {label} {
	if {([string compare [lindex $label 0] "i"] == 0) ||
	    ([string compare [lindex $label 0] "tau"] == 0)} {
		return 0
	} else {
		return 1
	}
}


# =========================================================================

proc todot {f name withenrs} {
	global aut
	set n enil
	puts $f "digraph \"$name\" {"
	puts $f "___all \[label=\"$name\",shape=box\]"
	set repr "___all -> S$aut(init) \["
	if {$withenrs} {
		append repr "label=\"$n\","
	}
	append repr "name=\"$n\"\]"
	puts $f $repr
	foreach t $aut(trans) {
		set s [lindex $t 0]
		set n [lindex $t 1]
		set l [lindex $t 2]
		set d [lindex $t 3]
		set repr "S$s -> S$d \[label=\""
		if {$withenrs} {
			append repr "$l\\n$n"
		} else {
			append repr $l
		}
		append repr "\",name=$n"
		if {![isvisible $l]} {
			append repr ",style=dotted"
		}
		append repr "\]"
		puts $f $repr
	}
	puts $f "};"	
}

proc tomatrix {f name} {
	global aut
	foreach t $aut(trans) {
		set s [lindex $t 0]
		set n [lindex $t 1]
		set l [lindex $t 2]
		set d [lindex $t 3]
		set nodes($s) 1
		set nodes($d) 1
		if {![info exists edge($s,$d)]} {
			set edge($s,$d) 1
		} else {
			incr edge($s,$d)
		}
	}

	if {[regexp {^[0-9]+$} $aut(init)]} {
		set n [lsort -integer [array names nodes]]
	} else {
		set n [lsort  [array names nodes]]
	}

	set firstx 1
	puts -nonewline {% [};
	foreach x $n {
		if {!$firstx} {
			puts -nonewline ","
		}
		set firstx 0
		puts -nonewline $x
	}
	puts {]}

	set firstx 1
	puts -nonewline {[};
	foreach x $n {
		if {!$firstx} {
			puts ","
		}
		set firstx 0
		puts -nonewline {[};
		set firsty 1
		foreach y $n {
			if {!$firsty} {
				puts -nonewline ", "
			}
			set firsty 0
			if {[info exists edge($x,$y)]} {
				puts -nonewline "$edge($x,$y)"
			} else {
				puts -nonewline "0"
			}
		}
		puts -nonewline {]};
	}
	puts {]};
}
