import java.io.*;
import java.util.*;
import lts.*;

public class TorxLTS {
	Automata automata;
	String [] modelAlphabet;
	// byte [] endState;

	String name;

	Vector transitions = new Vector();
	int nexttrans = 0;
	TorxTransition initial;

	Vector states = new Vector();
	int nextstate = 0;
	

	TorxLTS(Automata automata, String f) {
		this.automata=automata;
		this.name= extractname(f);
		
		modelAlphabet = automata.getAlphabet();
		// endState = automata.END();
		TorxState s = createState(automata.START());
		initial = createTrans(s);
		expand(automata);
	}

	private String extractname (String f) {
		int i, j;

		i = f.lastIndexOf("/");
		if (i >= 0)
			i++;
		else
			i = 0;
		if (f.endsWith(".lts"))
			j = f.lastIndexOf(".");
		else
			j = f.length();
		return f.substring(i, j);
	}

	public void expand(Automata automata) {
		String [] modelAlphabet = automata.getAlphabet();
		int i = 0;
		while (i < states.size()) {
			TorxState s = (TorxState)states.get(i);
			System.err.println("A_DEBUG expanding " + s.nr());
			s.explore_start();
			MyList succ = automata.getTransitions(s.ltsa());
			while (! succ.empty()) {
				s.explore_add(
				   createTrans(
					 s,
					 createState(succ.getTo()),
					 modelAlphabet[succ.getAction()]));
				succ.next();
			}
			s.explore_stop();
			i++;
		}
	}

	public void todot() {
		String s, l, d, n;
		int i;

		System.out.println("digraph \"" + name + "\" {");
		System.out.println("___all [label=\"" + name + "\",shape=box]");
		i = 0;
		while (i < transitions.size()) {
			TorxTransition t = (TorxTransition)transitions.get(i);
			if (t.src() == null) {
				s = "___all";
				l = "";
			} else {
				s = t.src().nr();
				l = t.label();
			}
			d = t.dst.nr();
			n = t.nr();
			System.out.println(s + " -> " + d + " [label=\"" + l + "\",name=" + n + "]");
			i++;
		}
		System.out.println("}");
	}
	
	// support for states

	public TorxState createState(byte[] s) {
		TorxState ts = new TorxState(s);
		if (!states.contains(ts)) {
			states.add(ts);
			ts.number = nextstate++;
			ts.ltsa = s;
		} else
			ts = lookupState(states.indexOf(ts));
		return ts;
	}
	
	public TorxState lookupState(int nr) {
		if (nr >=  states.size())
			return null;
		else
			return (TorxState)states.get(nr);
	}
	


	// support for transitions

	public TorxTransition initial() {
		return initial;
	}

	public TorxTransition createTrans(TorxState s, TorxState d, String l) {
		TorxTransition t = new TorxTransition(s, d, l, nexttrans++);
		transitions.add(t);
		return t;
	}
	
	// make sure that we create the root only once,
	// even if the user asks for the 'r' action multiple times.
	public TorxTransition createTrans(TorxState d) {
		if (d.incoming() == null)
			return createTrans(null, d, "<root>");
		else 
			return d.incoming();
	}
	
	public TorxTransition lookupTrans(int nr) {
		if (nr >=  transitions.size())
			return null;
		else {
			TorxTransition t = (TorxTransition)transitions.get(nr);
			if (t.dst().incoming() != t) {
				int id = t.dst().incoming().number();
				return (TorxTransition)transitions.get(id);
			} else
				return (TorxTransition)transitions.get(nr);
		}
	}
}
