# =========================================================================
#
# $Id: extension.tcl,v 1.1.2.4 2006/01/25 21:57:04 belinfan Exp $
#
# Author: Axel Belinfante
#
# This file implements the machinery underlying the PRIMER, GUIDE and MUTANT
# commands that we use in primer, guide and mutant config files.
# The interface to this machinery is via the interface functions in
# primers.tcl guides.tcl and mutants.tcl .
#
# The idea is to interpret primer, guide and mutant files as tcl files
# containing only invocations of the PRIMER, GUIDE resp. MUTANT command,
# to register primers, guides resp. mutants.
# As result of such registration, the GUI will add the registered
# items to the primers, guides resp. mutants menu.
# Question:
# will the loading of a second primer or guide or mutant file override the first one?
# Or can we have items of multiple primer/guide/mutant files 'active' at the same time?
# If so, how will this be visible in the menu?
# (as a sub-menu under the Primers/Guides/Mutants menu?)
# How will we be able to 'close' different primer/guide/mutant files?
# (via a sub-menu under the Close Primers/Guides/Mutants menu entry?)
#
# =========================================================================

# This routine will be invoked from xtorx to force auto-loading of this file.
proc do_first_initExtension {{arglist {}}} {
}

proc do_initExtension {kind pname wname mtitle
                       tvarname fvarname flagvarname cfgvarname
                       butenabled butdisabled
                       ftypes {arglist {}}} {
	global tester
	global debug

	# set debug(Msg,do_mutant_callback) 1
	# set debug(Msg,do_cleanupMutants) 1
	

	set tester(ext,$kind,procname) $pname
	set tester(ext,$kind,widgetname) $wname
	set tester(ext,$kind,titlename,varname) $tvarname
	set tester(ext,$kind,file,varname) $fvarname
	set tester(ext,$kind,torx,cmdline,flags) $flagvarname
	set tester(ext,$kind,torx,cmdline,config) $cfgvarname
	set tester(ext,$kind,butenabled) $butenabled
	set tester(ext,$kind,butdisabled) $butdisabled
	set tester(ext,$kind,open,types) $ftypes
	set tester(ext,$kind,open,mtitle) $mtitle

	set tester(ext,$kind,submenu) [WcFullNameToWidget . $wname]
	set tester(ext,$kind,nrof) 0
	set $tester(ext,$kind,file,varname) ""
	set $tester(ext,$kind,titlename,varname) ""
}

proc do_cleanupExtension {kind {arglist {}}} {
	global tester
	if {[info exists tester(ext,$kind,variable,prev)]} {
		# test to make sure that we indeed have a constructor name/string
		if {$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) != {}} {
			Debug_Msg "do_cleanupExtension" "$kind Destructor called for: $tester(ext,$kind,variable,prev)"
			if {[catch {eval "$tester(ext,$kind,destructor,$tester(ext,$kind,variable,prev)) $tester(ext,$kind,torx,cmdline,flags) $tester(ext,$kind,torx,cmdline,config)" } foo]} {
				catch {puts stderr "Error executing destructor for entry nr. $tester(ext,$kind,variable,prev) \"$tester(ext,$kind,entryname,$tester(ext,$kind,variable,prev))\": $foo"}
			}
		} else {
			Debug_Msg "do_cleanupExtension" "$kind No destructor for: $tester(ext,$kind,variable,prev)"
		}
		catch { unset tester(ext,$kind,variable,prev) }
	}
}

proc do_open_extension {kind {arglist {}}} {
	global tester
	if {[llength $arglist] >= 3} {
		set filename [tk_getOpenFile -filetypes $tester(ext,$kind,open,types) -initialdir $tester(open,initialdir) -title $tester(ext,$kind,open,mtitle) ]
		do_do_open_extension $kind $filename
	}
}
proc do_do_open_extension {kind filename} {
	global tester
	if {$filename != ""} {
		do_do_close_extension $kind
		set tester(ext,$kind,file) $filename
		set basename [file tail $filename]
		set $tester(ext,$kind,file,varname) [file rootname $basename]
		set tester(dirname) [file dirname $filename]
		set tester(open,initialdir) $tester(dirname)
		
		if {[update_extension_menu $kind $filename]} {
			update_title
		}
	}
}
proc update_extension_menu_add {kind file} {
	global tester
	do_cleanupExtension $kind
	# add entry 'none' and make it active
	# It will not have an associated radio button.
	# It will be represented by: no radiobuttons active
	# the user may add his own entry named none, if he wants to;
	# the user's 'none' entry _will_ have its associated radio button
	extension_register $tester(ext,$kind,nrof) $kind "none" "" {} {}
	set tester(ext,$kind,variable,cur) $tester(ext,$kind,nrof)
	set tester(ext,$kind,none) $tester(ext,$kind,nrof)
	
	if {[catch { source $file } foo ]} {
		catch {puts stderr "Error loading $kind file $file: $foo"}
		update_extension_menu_clear $kind
		return 0
	} else {
		WcSetCondSensitive   "" "" $tester(ext,$kind,butenabled)
		WcSetCondInSensitive "" "" $tester(ext,$kind,butdisabled)
		return 1
	}
}
proc update_extension_menu_clear {kind} {
	global tester
	do_cleanupExtension $kind
	$tester(ext,$kind,submenu) delete 0 end
	while {$tester(ext,$kind,nrof) > 0} {
		catch {unset $tester(ext,$kind,entryname,$tester(ext,$kind,nrof))}
		catch {unset $tester(ext,$kind,titlename,$tester(ext,$kind,nrof))}
		catch {unset $tester(ext,$kind,constructor,$tester(ext,$kind,nrof))}
		catch {unset $tester(ext,$kind,destructor,$tester(ext,$kind,nrof))}
		incr tester(ext,$kind,nrof) -1
	}
	set tester(ext,$kind,nrof) 0
	WcSetCondInSensitive "" "" $tester(ext,$kind,butenabled)
	WcSetCondSensitive   "" "" $tester(ext,$kind,butdisabled)
	return 1
}
proc update_extension_menu {kind {file ""}} {
	global tester
	if {$file != ""} {
		return [update_extension_menu_add $kind $file]
	} else {
		return [update_extension_menu_clear $kind]
	}
}
proc do_close_extension {kind {arglist {}}} {
	global tester
#	if {[llength $arglist] >= 3} {
		do_do_close_extension $kind
		update_title
#	}
}
proc do_do_close_extension {kind} {
	global tester
	set tester(ext,$kind,file) {}
	set $tester(ext,$kind,file,varname) {}
		
	# make sure that we unset the mutant when we close the file
	if {[info exists tester(ext,$kind,none)]} {
		set tester(ext,$kind,variable,cur) $tester(ext,$kind,none)
		do_extension_callback $kind
		update_extension_menu $kind
	}
}
proc do_reopen_extension {kind {arglist {}}} {
	global tester
	set filename $tester(ext,$kind,file)
	do_close_extension $kind $arglist
	update
	update idletasks
	do_do_open_extension $kind $filename
}

proc do_extension_callback {kind {arglist {}}} {
	global tester
	Debug_FnArgs "do_extension_callback" $kind $arglist
	if {[info exists tester(ext,$kind,variable,prev)]} {
		if {$tester(ext,$kind,variable,cur) == $tester(ext,$kind,variable,prev)} {
			return
		}
	}
	do_cleanupExtension $kind
	set tester(ext,$kind,variable,prev) $tester(ext,$kind,variable,cur)
	# test to make sure that we indeed have a constructor name/string
	if {$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) != {}} {
		Debug_Msg "do_extension_callback" "$kind Constructor called for: $tester(ext,$kind,variable,prev)"
		if {[catch {eval "$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) $tester(ext,$kind,torx,cmdline,flags) $tester(ext,$kind,torx,cmdline,config)" } foo]} {
			catch {puts stderr "Error executing constructor for entry nr. $tester(ext,$kind,variable,prev) \"$tester(ext,$kind,entryname,$tester(ext,$kind,variable,prev))\": $foo"}
		}
	} else {
		Debug_Msg "do_extension_callback" "$kind No constructor for: $tester(ext,$kind,variable,prev)"
	}
	set $tester(ext,$kind,titlename,varname) \
		$tester(ext,$kind,titlename,$tester(ext,$kind,variable,prev))
	update_title
}

proc do_do_select_extension {kind titlename} {
	global tester
	if {[info exists tester(ext,$kind,submenu)] &&
	    [$tester(ext,$kind,submenu) index end] > 0} {
		if {[catch {$tester(ext,$kind,submenu) invoke *[set titlename]*} foo]} {
		}
	} else {
		if {[string compare $titlename none] != 0} {
			set $tester(ext,$kind,titlename,varname) $titlename
		} else {
			set $tester(ext,$kind,titlename,varname) ""
		}
		update_title
	}
}

proc extension_register {key kind entry title create destruct {arglist {}}} {
	global tester
	set tester(ext,$kind,entryname,$key) $entry
	set tester(ext,$kind,titlename,$key) $title
	set tester(ext,$kind,constructor,$key) $create
	set tester(ext,$kind,destructor,$key) $destruct
	set tester(ext,$kind,args,$key) $arglist
}

proc EXTENSION {kind entry title create destruct {arglist {}}} {
	global tester
	Debug_FnArgs "EXTENSION" $kind $entry $title $create $destruct $arglist
	incr tester(ext,$kind,nrof)
	extension_register $tester(ext,$kind,nrof) $kind $entry $title $create $destruct $arglist
	
	$tester(ext,$kind,submenu) add radiobutton \
		-label $tester(ext,$kind,entryname,$tester(ext,$kind,nrof)) \
		-variable tester(ext,$kind,variable,cur) \
		-value $tester(ext,$kind,nrof) \
		-command "do_extension_callback $kind"

	# the name "none" is special: it will be the default setting
	if {[string compare $entry none] == 0} {
		set tester(ext,$kind,variable,cur) $tester(ext,$kind,nrof)
	}
}
