# =========================================================================
#
# start spec module and connect to it
# because we have several primers
# (effectively, zero or one spec,
#  and zero or more guidance information generators)
# we have to identify them. Together with the identification,
# we also have a 'role' (spec or guidance).
#

proc start_primer {id} {
	global timeout maininfo specinfo sidinfo
	match_max -d 15000
	set timeout $maininfo(timeout)
	if {[catch {spawn -pty} foo]} {
		send_error "spawn for stderr of $specinfo(role,$id) $id fails: $foo"
		do_cleanup
		exit 1
	}
	set sidinfo(spec,$id,stderr) $spawn_id
	set specinfo(pty,$id,stderr) $spawn_out(slave,name)
	send_error "A_DEBUG $specinfo(role,$id) $id Stderr match_max=[match_max]\n"

	generate_primer_command $id
#		if {$maininfo(debug) > 0} {
			send_error "DEBUG $id: $specinfo(cmd,$id)\n"
#		}
	
	if {[catch $specinfo(cmd,$id) foo]} {
		send_error "spawn for $specinfo(role,$id) $id fails: $foo\n"
		do_cleanup
		exit 1
	}
	set sidinfo(spec,$id) $spawn_id
	set specinfo(pid,$id) [exp_pid -i $sidinfo(spec,$id)]
	send_error "DEBUG $specinfo(role,$id) $id match_max=[match_max]\n"

	# we have to use 'eval' here, because we want the $id variable
	# to be expanded, because it will not exist in the context (scope)
	# in which this expect_after command will be executed.
	eval "
	expect_background {
		-i $sidinfo(spec,$id,stderr) full_buffer {
			send_error \"A_WARNING $specinfo(role,$id) $id Stderr Buffer Overflow!!!\\n\"
			set txt \$expect_out(buffer)
			send_error \"A_WARNING $specinfo(role,$id) $id Stderr Buffer Contents Skipped: \$txt\\n\"
			send_user \"A_ERROR $specinfo(role,$id) $id Stderr Buffer Contents Skipped: \$txt\\n\"
			exit
		}
		-i $sidinfo(spec,$id,stderr) -re \"^A_DEBUG\\\[ \\t]\\\[ \\t]*(\\\[^\\r\\n]*)\\r*\\n\" {
			set txt \$expect_out(1,string)
			send_error \"$specinfo(role,$id) $id Debug: \$txt\\n\"
		}
		-i $sidinfo(spec,$id,stderr) -re \"(\\\[^\\n\\r]*)\\r*\\n\" {
			set l \$expect_out(1,string)
			send_error \"$specinfo(role,$id) $id Stderr: \$l\\n\"
		}
	}
	"
	
	# we have to use 'eval' here, because we want the $id variable
	# to be expanded, because it will not exist in the context (scope)
	# in which this expect_after command will be executed.
	# NOTE: we CAN NOT split this big expect_after command in several
	# single expect_after commands, because each new expect_after command
	# issued with the same "-i spawnid" flag may replace (and thus disable)
	# the previous expect_after command set with the same "-i spawnid" flag.
	
	eval "
	expect_before {
		-i $sidinfo(spec,$id) full_buffer {
			send_error \"Warning $specinfo(role,$id) $id: Buffer Overflow!!!\\n\"
			set txt \$expect_out(buffer)
			send_error \"Warning $specinfo(role,$id) $id: Buffer Contents Skipped: \$txt\\n\"
			sleep 1
			if {\[info exists sidinfo(spec,$id)]} {
				unset sidinfo(spec,$id)
			}
			catch {fatal \"Warning $specinfo(role,$id) $id: Buffer Contents Skipped: \$txt\\n\"}
			exit
		}
	}
	expect_after {
		-i $sidinfo(spec,$id) -re \"^C_\\\[^\\n\\r]*\\r*\\n\" {
			# skip. This is echo'ed back from the specification module
		}
		-i $sidinfo(spec,$id) \"^A_QUIT\\\[ \\t]*\\r*\\n\" {
			set result [list error]
		}
		-i $sidinfo(spec,$id) -re \"^A_WARNING(\\\[^\\r\\n]*)\\r*\\n\" {
			set txt \$expect_out(1,string)
			send_error \"Warning $specinfo(role,$id) $id: \$txt\\n\"
		}
		-i $sidinfo(spec,$id) -re \"^A_LOG(\\\[^\\r\\n]*)\\r*\\n\" {
			set txt \$expect_out(1,string)
			send_user \"A_LOG \$txt\\n\"
		}
		-i $sidinfo(spec,$id) -re \"^A_DEBUG(\\\[^\\r\\n]*)\\r*\\n\" {
			set txt \$expect_out(1,string)
			send_error \"Debug $specinfo(role,$id) $id: \$txt\\n\"
		}
		-i $sidinfo(spec,$id) -re \"^A_ERROR UnknownCommand(\\\[^\\r\\n]*)\r*\\n\" {
			set txt \$expect_out(1,string)
			send_error \"Warning $specinfo(role,$id) $id: unimplemented: \$txt\\n\"
			set result [list error]
		}
		-i $sidinfo(spec,$id) -re \"^A_ERROR(\\\[^\\r\\n]*)\r*\\n\" {
			set txt \$expect_out(1,string)
			fatal \"Error $specinfo(role,$id) $id: \$txt\\n\"
		}
		-i $sidinfo(spec,$id) -re \"^(A_\\\[^\\n\\r]*)\\r*\\n\" {
			set txt \$expect_out(1,string)
			send_error \"Unexpected $specinfo(role,$id) $id: \$txt\\n\"
			set result [list error ]
		}
		-i $sidinfo(spec,$id) -re \"^(\\\[^\\n]*)\\r*\\n\" {
			set l \$expect_out(buffer)
			regexp -- \"^(\\\[^\\n\\r]*)\" \$l match sub
			send_error \"\$specinfo(role,$id) $id: \$sub\\n\"
			exp_continue
		}
		-i $sidinfo(spec,$id) eof {
			global sidinfo specinfo
			if {\[info exists sidinfo(spec,$id)]} {
				unset sidinfo(spec,$id)
			}
			if {\[info exists specinfo(pid,$id)]} {
				unset specinfo(pid,$id)
			}
			set index \[lsearch -exact \$specinfo(primerids) $id]
			if {\$index >= 0} {
				set specinfo(primerids) \
				    \[lreplace \$specinfo(primerids) \$index \$index]
			}
			set index \[lsearch -exact \$specinfo(specids) $id]
			if {\$index >= 0} {
				set specinfo(specids) \
				    \[lreplace \$specinfo(specids) \$index \$index]
			}
			set index \[lsearch -exact \$specinfo(guideids) $id]
			if {\$index >= 0} {
				set specinfo(guideids) \
				    \[lreplace \$specinfo(guideids) \$index \$index]
			}
			if {!\[info exists specinfo(quit,$id)]} {
				fatal \"End-of-file from \$specinfo(role,$id) $id\\n\"
			}
		}
	}
	"
}

# =========================================================================
proc start_primers {} {
	global specinfo 
	foreach id $specinfo(primerids) {
		start_primer $id
		spec_get_stats $id
	}
}
