# =========================================================================
# $Id: main_loop.tcl,v 1.1.2.13 2000/11/08 11:05:50 belinfan Exp $
# =========================================================================

proc primer_interface {li lu seed delta file susp} {
	if {[catch {do_primer_interface $li $lu $seed $delta $file $susp} foo]} {
		bgerror $foo
	}
}
proc do_primer_interface {li lu seed delta file susp} {
	primer_init $li $lu $delta $file $susp
	RandomInit $seed
	install_pending_events_handler
	disable_selection_exportation
	primer_start ;# compute initial menu
	restore_selection_exportation
	primer_main
}

proc primer_main {} {
	## Currently, we prefere to use blocking IO
	## fconfigure -nonblocking stdin
	fileevent stdin readable primer_try_handle_line	
}

# ====================================

proc disable_selection_exportation {} {
	global selexp_stack
	lappend selexp_stack [export_selection 0]
}
proc restore_selection_exportation {} {
	global selexp_stack
	set old [lindex $selexp_stack end]
	set selexp_stack [lreplace $selexp_stack end end]
	update idletasks ;# make selections before enabling export
	export_selection $old
}


# ====================================

proc get_arguments {arg} {
	global env
	# we expect the seed in env var SEED.
	# we expect the delta-event representation in env var DELTA.
	# we expect the config file name in env var CONFIGFILE.
	set result ""
	switch -- [string tolower $arg] {
		  li {
		  	set result [split [string trim $env(INGATES)] ,]
		} lu {
			set result [split [string trim $env(OUTGATES)] ,]
		} seed {
			set result [string trim $env(SEED)]
		} delta {
			set result [string trim $env(DELTA)]
		} config {
			set result [string trim $env(CONFIGFILE)]
		} suspension {
			set result [string trim $env(SUSPENSION)]
		}
	}
	return $result
}

# ====================================

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc bgerror {msg} {
	global errorInfo
	debugMsg "Fatal InternalError: $msg"
	set lines [split $errorInfo \n]
	debugMsg "Stack Begin"
	foreach l $lines {
		debugMsg "Stack: $l"
	}
	debugMsg "Stack End"
	# exit
}


# ====================================

proc install_pending_events_handler {} {
	after idle do_handle_pending_events
}

proc stop_pending_events_handler {} {
	global pending_events_handler_signal
	catch { unset pending_events_handler_signal }
}

proc do_handle_pending_events {} {
	global pending_events_handler_signal
	# debugMsg do_handle_pending_events
	handle_pending_events
	if {[info exists pending_events_handler_signal]} {
		after 1 install_pending_events_handler
	}
}

# ====================================

proc primer_try_handle_line {} {
	fileevent stdin readable ""
	set stop 0
	set skip 0
	set line ""
	if {[gets stdin line] < 0} {
		if {[fblocked stdin]} {
			set skip 1
		} elseif {[eof stdin]} {
			set line "C_QUIT"
			set stop 1
		}
	}
	if {! $skip} {
		disable_selection_exportation
		set speed [time {primer_handle_line $line}]
		puts "A_DEBUG time: $speed"
		restore_selection_exportation
	}
	if {! $stop} {
		fileevent stdin readable primer_try_handle_line
	}
}

# ====================================

proc primer_handle_line {line} {
	# debugMsg "Line: ($line)"
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [primer_parse $line]
		set parsed_ok [lindex $parsed_line 0]
		debugMsg "Parsed: ($parsed_line)"
		if {$parsed_ok} {
			set parsed_cmd [lindex $parsed_line 1]
			set cmd_keyword [lindex $parsed_cmd 0]
			switch -- [string toupper $cmd_keyword] {
			    C_QUIT {
					send_menu_verdict
			    	set cont 0
			    	puts "A_QUIT"
			  } C_CHANNELS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set kind_sug [field_value $fields iokind]
			  		primer_do_channels $kind_sug
			  } C_IOKIND {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kind_sug [field_value $fields iokind]
			  		primer_do_iokind $kind_sug $chan_sug
			  } C_INPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		primer_do_input $ev $chan $s
			  } C_OUTPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		primer_do_output $ev $chan $s
			  } C_INPUTS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kinds [list input inputsuspension]
			    	primer_menu "A_EVENT" "INPUTS" $kinds $chan_sug
			  } C_OUTPUTS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kinds [list output outputsuspension]
			    	primer_menu "A_EVENT" "OUTPUTS" $kinds $chan_sug
			  } C_STATS {
					send_menu_verdict
			    	puts "A_STATS no stats yet"
			  } default {
					puts "A_ERROR UnknownCommand: $cmd_keyword"
			  }
			}
		} else {
			puts "A_ERROR UnknownCommand: $line"
		}
	}
}

# =========================================================================
