# =========================================================================
# $Id: main_util.tcl,v 1.1.2.14 2000/09/13 13:24:06 belinfan Exp $
# =========================================================================

proc primer_init {li lu delta file suspension} {
	global susp_history
	set susp_history(input) {}
	set susp_history(output) {}
	event_kind_init $li $lu $delta $file $suspension
	reset_menu_verdict
}

proc primer_start {} {
	simulate [specification_name]
	obs_menu 0
}

proc primer_menu {prefix keyword kinds chan_sug} {
	set chan_opt {}
	if {[string compare $chan_sug ""] != 0} {
		if {![is_channel $chan_sug]} {
			puts "A_ERROR WrongValue: not a channel: \"$chan_sug\""
			return
		} elseif {[lsearch $kinds [get_kind $chan_sug]] < 0} {
			puts "A_ERROR Inconsistency: wrong iokind (not $keyword) channel: \"$chan_sug\""
			return
		} else {
			lappend chan_opt $chan_sug
		}
	}
	puts "A_[set keyword]_START"
	foreach el [mr_list $kinds $chan_opt] {
		set kind [lindex $el 0]
		set chan [lindex $el 1]
		set ev [smile_event2event [lindex $el 2]]
		switch -- $kind {
			  inputsuspension -
			  outputsuspension {
				set s_attr "\tsuspension=1"
			} input -
			  output {
				set s_attr "\tsuspension=0"
			} default {
				set s_attr ""
			}
		}
		set c_attr "channel=$chan"
		puts "$prefix event=$ev$s_attr\t$c_attr"
	}
	puts "A_[set keyword]_END"
}

proc primer_parse {line} {
	# debugMsg "ToParse: ($line)"
	if {[regexp {^[ \t]*([cC]_[a-zA-Z0-9_]+)([ \t].*)?} $line match cmd rest]} {
		set the_rest [string trim $rest]
		return [list 1 [list $cmd $the_rest]]
	} else {
		return [list 0]
	}
}

proc primer_do_input {ev chan s} {
	return [primer_do_next INPUT input $ev $chan $s]
}
proc primer_do_output {ev chan s} {
	return [primer_do_next OUTPUT output $ev $chan $s]
}

proc primer_do_next {keyword iokind ev chan s} {
	global susp_history
	set comb [primer_combine_fields $iokind $ev $chan $s C_[set keyword]]
	if {[lindex $comb 0]} {
		set kind [lindex $comb 1]
		set ev [lindex $comb 2]
		set chan [lindex $comb 3]
		set s [lindex $comb 4]
		if {$s} {
			set trying_suspension $chan
			set res [primer_next [set iokind]suspension $ev $chan]
		} else {
			set trying_suspension ""
			set res [primer_next [set iokind] $ev $chan]
		}
		if {[llength $res] > 0} {
			obs_menu $res
			if {[string compare $trying_suspension ""] != 0} {
				if {[catch {suspension_representation $chan} res_ev]} {
					set res_ev ""
					set s_attr ""
				} else {
					set s_attr "\tsuspension=1"
					if {[lsearch $susp_history($iokind) $trying_suspension] < 0} {
						lappend susp_history($iokind) $trying_suspension
					}
				}
			} else {
				set res_ev [smile_event2event [show_event [lindex $res 0]]]
				set s_attr "\tsuspension=0"
				set susp_history($iokind) {}
			}
			if {[catch {event_kind $res_ev} kind_chan]} {
				puts "A_ERROR UnknownIOKind: event=$res_ev"
				return
			}
			set c [lindex $kind_chan 1]
			set c_attr "channel=$c"
			puts "A_[set keyword]_OK event=$res_ev$s_attr\t$c_attr"
		} else {
			puts "A_[set keyword]_ERROR"
		}
	} else {
		puts [lindex $comb 1]
	}
}

# primer_combine_fields:
# combines the given fields; return either
#    [list 1 kind ev chan s] (succesfull combined fields)
#    [0 errmsg]              (unsufficient or incomplete info)
#
# Assumptions:
#  - we either generate or recognize suspension-events ((m)ioco resp traces)
#  - suspension-events may have an empty-string representation;
#    this should not pose problems, because we can describe them
#    using fields chan and s.
#  - each given field may be emtpy string; this only poses problems
#    if the other fields do not contain sufficient information
#    to have a result.
proc primer_combine_fields { kind ev chan s cmd } {
	switch -- $kind {
		input -
		output {
		} default {
			return [list 0 "A_ERROR InternalError: primer_combine_fields: kind notin (input,output): \"$kind\""
		}
	}
	if {[string compare $chan ""] != 0} {
		if {[is_channel $chan]} {
			set channel $chan
		} else {
			return [list 0 "A_ERROR WrongValue: not a channel: \"$chan\""]
		}
	}
	if {[string compare $s ""] != 0} {
		if {$s} {
			set suspension 1
		} else {
			set suspension 0
		}
	}
	if {([string compare $ev ""] == 0)} {
		if {([info exists suspension] &&
		     [info exists channel])} {
			if {$s} {
				if {[catch {suspension_representation $channel} ev]} {
					set ev ""
				} else {
					set ev_is_suspension 1
					set ev_kind [set kind]suspension
					set ev_channel $channel
				}
			}
		}
	} else {
		if {[catch {event_kind $ev} kind_chan_given]} {
			return [list 0 "A_ERROR UnknownIOKind: event=$ev"]
		}
		set ev_kind  [lindex $kind_chan_given 0]
		set ev_channel [lindex $kind_chan_given 1]
		switch -- $ev_kind {
			input -
			output {
				set ev_is_suspension 0
			} inputsuspension -
			  outputsuspension {
				set ev_is_suspension 1
			}
		}
	}
	if {[info exists channel] &&
	    ![info exists ev_is_suspension]} {
		if {![catch {suspension_representation $channel} drepr ]} {
			if {[eq_event $ev $drepr]} {
				set ev_is_suspension 1
				set ev_kind [set kind]suspension
			} else {
				set ev_is_suspension 0
			}
		} else {
			set ev_is_suspension 0
		}
	}
	if {[info exists suspension] &&
	    [info exists ev_is_suspension]} {
		if {$ev_is_suspension != $suspension} {
			return [list 0 "A_ERROR Inconsistency: event and suspension fields contradict each other"]
		}
	}
	if {[info exists channel] &&
	    [info exists ev_channel]} {
		if {[string compare $channel $ev_channel] != 0} {
			return [list 0 "A_ERROR Inconsistency: event and channel fields contradict each other"]
		}
	}
	if {[info exists ev_channel] &&
	    [info exists ev_kind] &&
	    [info exists ev_is_suspension]} {
		if {$ev_is_suspension} {
			return [list 1 $ev_kind $ev $ev_channel $ev_is_suspension]
		} elseif {[string compare $ev ""] != 0} {
			return [list 1 $ev_kind $ev $ev_channel $ev_is_suspension]
		} else {
			return [list 0 "A_ERROR ArgumentMissing: Usage: $cmd event-to-be-checked"]
		}
	} else {
		return [list 0 "A_ERROR ArgumentMissing: Usage: $cmd event-to-be-checked"]
	}
}

proc primer_do_iokind {kind_sug chan_sug} {
	global susp_history
	# debugMsg "primer_do_iokind ($kind_sug) ($chan_sug)"
	if {[string compare $kind_sug ""] != 0} {
		if {![is_iokind $kind_sug]} {
			set result "A_ERROR WrongValue: not an iokind: \"$kind_sug\""
		}
	}
	if {[string compare $chan_sug ""] != 0} {
		if {![is_channel $chan_sug]} {
			set result "A_ERROR WrongValue: not a channel: \"$chan_sug\""
		}
	}
	if {![info exists result]} {
		if {([string compare $kind_sug ""] != 0) &&
		    ([string compare $chan_sug ""] != 0)} {
		    if {[string compare $kind_sug [get_kind $chan_sug]] != 0} {
				set result "A_ERROR Inconsistency: channel and iokind fields conflict"
		    } else {
				set result "A_IOKIND iokind=$kind_sug\tchannel=$chan_sug"
			}
		} elseif {[string compare $kind_sug ""] != 0} {
			set chan [primer_choose_channel $kind_sug]
			set result "A_IOKIND iokind=$kind_sug\tchannel=$chan"
		} elseif {[string compare $chan_sug ""] != 0} {
			set kind [primer_choose_kind $chan_sug]
			set result "A_IOKIND iokind=$kind\tchannel=$chan_sug"
		} else {
			if {![primer_menu_contains [list input]]} {
				set chan [primer_choose_channel output]
				set result "A_IOKIND iokind=output\tchannel=$chan"
			} elseif {[llength $susp_history(output)] ==
			          [llength [get_channels outputsuspension]]} {
				set chan [primer_choose_channel input]
				set result "A_IOKIND iokind=input\tchannel=$chan"
			} else {
				# FIXME: we still may have to avoid suspension on one or more channels!
				# we keep a list of channels on which we
				# have recently seen suspension!
				# randomly choose between input and output
				set kind [primer_choose_kind ]
				set chan [primer_choose_channel $kind]
				set result "A_IOKIND iokind=$kind\tchannel=$chan"
			}
		}
	}
	debugMsg "primer_do_iokind ($kind_sug) ($chan_sug): $result"
	puts $result
}

proc primer_do_channels {kind_sug} {
	if {[string compare $kind_sug ""] != 0} {
		if {![is_iokind $kind_sug]} {
			set result "A_ERROR WrongValue: not an iokind: \"$kind_sug\""
		} else {
			set kinds [list $kind_sug ]
		}
	} else {
		set kinds [list input output]
	}
	if {![info exists result]} {
		puts "A_CHANNELS_START"
		foreach k $kinds {
			set cl [get_channels $k]
			foreach c $cl {
				puts "A_CHANNEL iokind=$k\tchannel=$c"
			}
		}
		puts "A_CHANNELS_END"
	} else {
		puts $result
	}
}


proc primer_menu_contains {kinds} {
	foreach k $kinds {
		if {[llength [mr_list [list $k]]] > 0} {
			return 1
		}
	}
	return 0
}

proc primer_choose_channel { {kind ""} } {
	if {[string compare $kind ""] == 0} {
		set c [concat [get_channels input] [get_channels output]]
	} else {
		set c [get_channels $kind]
	}
	set l [llength $c]
	set r [random $l]
	return [lindex $c $r]
}

proc primer_choose_kind { {chan ""} } {
	if {[string compare $chan ""] == 0} {
		if {[random 2]} {
			set result input
		} else {
			set result output
		}
	} else {
		return [get_kind $chan]
	}
}

proc smile_event2event {ev} {
	set tmp ""
	append tmp [lindex $ev 0] ;# the gate name
	set exps [lindex $ev 1]
	set preds [lindex $ev 2]
	foreach exp $exps {
		set question_exclam [lindex $exp 0] ;# the ! or ?
		if {[string compare $question_exclam "?"] == 0} {
			append tmp " !"
		} else {
			append tmp " $question_exclam"
		}
		append tmp " [lindex $exp 1]" ;# the expression
	}
	return $tmp
}

proc send_menu_verdict {} {
	global menuinfo
	if {[string compare $menuinfo(verdict) ""] != 0} {
		puts "A_GUIDANCE_VERDICT $menuinfo(verdict)"
	} elseif {$menuinfo(inconsistent)} {
		puts "A_GUIDANCE_VERDICT INCONSISTENT"
	}
}

proc set_menu_verdict {s} {
	global menuinfo
	lappend menuinfo(verdict) $s
}

proc reset_menu_verdict {} {
	global menuinfo
	set menuinfo(verdict) {}
	set menuinfo(inconsistent) 0
}
