# =========================================================================
# $Id: menu.tcl,v 1.1.2.21 2000/09/13 13:24:11 belinfan Exp $
# =========================================================================

proc obs_menu { S } {
	mr_empty
	reset_menu_verdict
	rec_obs_menu $S
	return [mr_events]
}

# ==========================================

proc rec_obs_menu { S } {
	foreach s $S {
		expand_state $s
	}
}

proc expand_state { s } {
	# debugMsg "expanding $s"
	set events [explorer_expand $s]
	debugMsg "expanded $s: $events"
	foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
		set add_suspension($c) 1
		set suspension_predicates($c) {}
	}
	foreach e $events {
		set repr [show_event $e]
		set ev_printable [smile_event2event $repr]
		if {[catch {event_kind $ev_printable} kind_chan ]} {
			puts "A_WARNING UnknownIOKind: event=$ev_printable"
			continue
		}
		set kind [lindex $kind_chan 0]
		set chan [lindex $kind_chan 1]
		debugMsg "visibility of $e: $kind $chan"
		switch -- $kind {
			  input -
			  output {
			  	if {[predicates_solved $e]} {
					set add_suspension($chan) 0
				} else {
					set preds [predicates $e]
			  		warningMsg "Predicates Unsolved for $kind $e: $preds"
					lappend suspension_predicates($chan) $preds
				}
				mr_insert $kind $chan $e $repr
			} inputsuspension -
			  outputsuspension {
				if {[config_suspension_detection $kind]} {
			  		warningMsg "menu contains explicit suspension event for channel: \"$chan\""
				} else {
			  		if {[predicates_solved $e]} {
						set add_suspension($chan) 0
					} else {
						set preds [predicates $e]
			 	 		warningMsg "Predicates Unsolved for $kind $e: $preds"
						lappend suspension_predicates($chan) $preds
					}
					mr_insert $kind $chan $e $repr
				}
			} internal {
			  	if {[predicates_solved $e]} {
			  		foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
						set add_suspension($c) 0
					}
				} else {
					set preds [predicates $e]
			  		warningMsg "Predicates Unsolved for $kind $e: $preds"
			  		foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
						lappend suspension_predicates($chan) $preds
					}
				}
				set identical_to_expanded [mr_insert $kind $chan $e $repr]
				if {! $identical_to_expanded} {
					rec_obs_menu [list $e]
				} else {
					debugMsg "not expanding identical $kind $e"
				}
			} verdict {
				set_menu_verdict $chan
			}
		}
	}
	foreach k [list inputsuspension outputsuspension] {
		foreach c [get_channels $k] {
			if {$add_suspension($c)} {
				if {[config_suspension_detection $k]} {
					if {[catch {list [suspension_representation $c] {} {} } repr]} {
						set repr [list "" {} {} ]
					}
					# mr_insert $k $c $s $repr $suspension_predicates
					mr_insert $k $c $s $repr
				} else {
			  		warningMsg "empty menu (without suspension event) for channel: \"$c\""
				}
			}
		}
	}
}

proc explorer_expand { s } {
	global unfold_depth
	global rewrite_always
	global rewrite_minimal
	global unparse_incremental
	set unfold_depth 1
	set rewrite_always 1
	set rewrite_minimal 1
	# set unparse_incremental 0
	return [unfold $s]
}

proc event_kind {repr} {
	global gates
	set label [string tolower $repr]
	set l [split $label]
	# debugMsg "event_kind: searching ($repr)($label)"
	if {[string compare [lindex $l 0] "i"] == 0} {
		return [list internal ]
	} else {
		set ok_kind_chan [em_find_string $label]
		if {[lindex $ok_kind_chan 0]} {
			return [lindex $ok_kind_chan 1]
		} else {
			return -code error -errorinfo "unknown event kind"
		}
	}
}

proc event_kind_init {li lu delta file susp} {
	global gates
	global delta_event_repr

	config_init $li $lu $delta $susp
	config_from_file $file
	config_finish
}

proc suspension_representation {chan} {
	return [config_get_suspension_representation $chan]
}

proc get_channels { kind } {
	return [config_get_channels $kind]
}

proc get_kind { channel } {
	return [config_get_iokind $channel]
}

proc is_channel { chan } {
	return [config_is_channel $chan]
}

proc is_iokind { kind } {
	return [config_is_iokind $kind]
}
