# =========================================================================
# $Id: menurepr.tcl,v 1.1.2.10 2000/09/12 15:48:26 belinfan Exp $
# =========================================================================

## Build a menu structure.
## This menu is a list of lists, where each sublist
## contains menu-elements with 'identical' events.
## The interface to this structure is formed by the
## a the following routines.
## Currently implemented actions:
proc mr_init {} {
	global mr
	catch {
		unset mr
	}
	set mr(kinds) [list input output inputsuspension outputsuspension internal verdict]
	foreach k [concat $mr(kinds)] {
		set mr(channels,$k) [get_channels $k]
		if {[llength $mr(channels,$k)] <= 0} {
			set mr(channels,$k) [list ""]
		}
		foreach c $mr(channels,$k) {
			set mr(keys,$k,$c) {}
			set mr(nextkey,$k,$c) 1
		}
	}
}

proc mr_empty {} {
	mr_init
}

#
# returns 1 if identical element is already in the menu
# (i.e. the inserted element need not be expanded);
# otherwise, returns 0.
proc mr_insert { kind chan event repr } {
	global mr
	set result 0
	set elsewhere ""
	if {![catch { set elsewhere [identical_node $event] } ]} {
		set mr(identical,$event) $elsewhere
		if {[info exists mr(event2sublist,$elsewhere)]} {
			debugMsg "identical $kind $event: $elsewhere (in-menu)"
			mr_do_insert $kind $chan $event $repr -elsewhere
			set result 1
		} else {
			debugMsg "identical $kind $event: $elsewhere (unseen)"
			mr_do_insert $kind $chan $event $repr
			mr_do_insert $kind $chan $elsewhere "" -elsewhere-unseen
		}
	} else {
		mr_do_insert $kind $chan $event $repr
	}
	return $result
}

#proc mr_identical { e } {
#	global mr
#	if {[info exists mr(identical,$e)]} {
#		return $mr(identical,$e)
#	} else {
#		return -1
#	}
#}

proc mr_in_menu { e } {
	global mr
	if {[info exists mr(identical,$e)]} {
		return 1
	} else {
		return 0
	}
}

proc mr_events { {kind_opt ""} } {
	global mr
	set events {}
	if {[string compare $kind_opt ""] != 0} {
		set kind $kind_opt
		foreach chan $mr(channels,$kind) {
			foreach key $mr(keys,$kind,$chan) {
				lappend events $mr(sublist,$kind,$chan,$key)
			}
		}
	} else {
		foreach kind $mr(kinds) {
			foreach chan $mr(channels,$kind) {
				foreach key $mr(keys,$kind,$chan) {
					lappend events $mr(sublist,$kind,$chan,$key)
				}
			}
		}
	}
	return $events
}
proc mr_show {} {
	return [join [mr_list] "\n"]
}
proc mr_display {} {
	return [mr_show]
}
proc mr_list { {kinds_opt {}} {chan_opt {}} } {
	global mr
	set events {}
	if {[llength $kinds_opt] > 0} {
		set kinds {}
		foreach kind $mr(kinds) {
			if {[lsearch -exact $kinds_opt $kind] >= 0} {
				lappend kinds $kind
			}
		}
	} else {
		set kinds $mr(kinds)
	}
	foreach kind $kinds {
		if {[llength $chan_opt] > 0} {
			set channels {}
			foreach chan $mr(channels,$kind) {
				if {[lsearch -exact $chan_opt $chan] >= 0} {
					lappend channels $chan
				}
			}
		} else {
			set channels $mr(channels,$kind)
		}
		foreach c $channels {
			set events [concat $events [mr_list_kind_chan $kind $c]]
		}
	}
	return $events
}

# ===========================================================

proc mr_do_do_insert {kind chan e repr elsewhere key} {
	global mr
	lappend mr(sublist,$kind,$chan,$key$elsewhere) $e
	set mr(event2sublist,$e) [list $kind $chan $key$elsewhere]
	set mr(repr,$kind,$chan,$e) $repr
	if {[string compare $elsewhere ""] == 0} {
		set mr(freevars,$kind,$chan,$e) [free_variables $e]
	}
}

proc mr_do_insert { kind chan e repr {elsewhere ""} } {
	global mr
	switch -- $kind {
		inputsuspension -
		outputsuspension -
		verdict {
			foreach key $mr(keys,$kind,$chan) {
				mr_do_do_insert $kind $chan $e $repr $elsewhere $key
				return
			}
			# if we make it to here, we did not find a submenu for this kind.
			if {[string compare $elsewhere ""] == 0} {
				# So, we create a new sublist.
				lappend mr(keys,$kind,$chan) $mr(nextkey,$kind,$chan)
				mr_do_do_insert $kind $chan $e $repr $elsewhere $mr(nextkey,$kind,$chan)
				incr mr(nextkey,$kind,$chan)
			} else {
				mr_do_do_insert $kind $chan $e $repr $elsewhere -1
			}
		} default {
			foreach key $mr(keys,$kind,$chan) {
				# try to match the element with one
				# element of the sublist.
				# NOTE: We assume here that each
				# element of the sublist can be
				# taken as a representative of it,
				# i.e. it doesn't matter with which
				# one we compare the element.
				# So, we take the first one.
				if {[match $e [lindex $mr(sublist,$kind,$chan,$key) 0]]} {
					mr_do_do_insert $kind $chan $e $repr $elsewhere $key
					return
				}
			}
			# if we make it to here, we did not find a match.
			if {[string compare $elsewhere ""] == 0} {
				# So, we create a new sublist.
				lappend mr(keys,$kind,$chan) $mr(nextkey,$kind,$chan)
				mr_do_do_insert $kind $chan $e $repr $elsewhere $mr(nextkey,$kind,$chan)
				incr mr(nextkey,$kind,$chan)
			} else {
				mr_do_do_insert $kind $chan $e $repr $elsewhere -1
			}
		}
	}
}

proc mr_list_kind_chan { kind chan } {
	global mr
	set events {}
	foreach key $mr(keys,$kind,$chan) {
		switch -- $kind {
			inputsuspension -
			outputsuspension {
				set representative [lindex $mr(sublist,$kind,$chan,$key) 0]
				set repr $mr(repr,$kind,$chan,$representative)
			} input -
			  output {
				set representative [lindex $mr(sublist,$kind,$chan,$key) 0]
				set free_vars $mr(freevars,$kind,$chan,$representative)
				set repr $mr(repr,$kind,$chan,$representative)
				foreach fv $free_vars {
					regsub [lindex $fv 0] $repr "var_[lindex $fv 1]" repr
				}
			} default {
				# handle verdicts???
			}
		}
		set ext_sublist {}
		foreach ext { -elsewhere  -elsewhere-unseen } {
			if {[info exists mr(sublist,$kind,$chan,$key$ext)]} {
				lappend ext_sublist $mr(sublist,$kind,$chan,$key$ext)
			} else {
				lappend ext_sublist {}
			}
		}
		lappend events [list $kind $chan $repr $mr(sublist,$kind,$chan,$key) $ext_sublist]
	}
	return $events
}
