#! @EXPECT@ --

# this could be run by tclsh instead of expect
# the only reason we use expect here is that we look for expect
# in the configure script anyway 

# translate lineair form of test case tree generated by
#   torx --suite
# to .aut file
#
# we read multiple lines from stdin and
# for each line we generate an .aut file on stdout

# input lines are of the following form:

# S ; 1 ; button ; S ; 1 ; button ; S ; 1 ; button ; O ; 3 ; cappuccino ; S ; 1 ; button ; P ; 0 ; espresso ; O ; 1 ; delta ; P ; 0 ; coffee ; P ; 0

proc mkstate {{S {}} {d {}} {kind {}}} {
	global states

	if {![info exists states(next)]} {
		set states(next) 0
		set states(start) $states(next)
	}
#	set key "$S,$d,$kind"
	set key $states(next)
	if {![info exists states(map,$key)]} {
		set states(map,$key) $states(next)
		set states(next) [expr $states(next) + 1]
	}
	return $states(map,$key)
}

proc mktrans {src t dst} {
	global trans

	lappend trans($src) [list $t $dst]
}

proc writeaut {s} {
	global states
	global trans
	
	set nrstates [expr [llength [array names states]] - 2]
	set nrtrans 0
	foreach s [array names trans] {
		incr nrtrans [llength $trans($s)]
	}
	puts "des ($states(start), $nrtrans, $nrstates)"
	foreach s [lsort -integer [array names trans]] {
		foreach t $trans($s) {
			puts "($s, \"[lindex $t 0]\", [lindex $t 1])"
		}
	}
}
proc resetaut {} {
	global states
	global trans
	
	catch { unset states }
	catch { unset trans }
}

proc linstr2aut { s } {
	set arr [split $s \;]
	resetaut
	set s [mkstate]
	set arr [lin2aut $arr $s]
	writeaut $s
}

proc lin2aut { arr s } {
	if {[llength $arr] <= 0} {
		return
	}
	set hd [string trim [lindex $arr 0]]
	set arr [lrange $arr 1 end]
#	puts stderr "lin2aut doing $hd  ($arr)"
	switch $hd {
	  S -
	  O -
	  P {
		if {[llength $arr] <= 0} {
			return
		}
	  	set nsub [lindex $arr 0]
	  	if {[string compare $nsub ""] == 0} {
	  		set nsub 0
	  	}
		if {[llength $arr] <= 0} {
			return
		}
		set arr [lrange $arr 1 end]
		set i 0
		while {$i < $nsub} {
		  	set t [string trim [lindex $arr 0]]
			set arr [lrange $arr 1 end]
#			puts stderr "doing sub $i of $nsub \"$t\""
			set sprime [mkstate]
			mktrans $s $t $sprime
			set arr [lin2aut $arr $sprime]
			incr i
		}
		return $arr
	  }
	}
}

while {[gets stdin line] >= 0} {
	linstr2aut $line
}
