#! @TCLSH@
#! @EXPECT@ --
#! /Utils/bin/tclsh
#! /usr/local/bin/expect --

# NOTE: this script only needs tcl(sh); it does not use expect
# features. The only reason that we use the expect interpreter
# here is that the configure script of TorX finds the location of
# expect already for us, and (AFAIK) does not (yet?) look for tclsh.

# however,
# seems this works better with tclsh than with expect on windows

proc find_user {} {
	global env
	global tcl_platform

	if {[info exists tcl_platform(user)]} {
		set user $tcl_platform(user)
	} elseif {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} elseif {[info exists env(USERNAME)]} {
		set user $env(USERNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_tmp {} {
	global env
	return /tmp
	# Stupid temp directory
	set S(tmp) [pwd]
	if {[file isdirectory "c:/temp"]} { set S(tmp) "c:/temp"}
	if {[file isdirectory "/tmp"]} { set S(tmp) "/tmp"}
	catch {set S(tmp) $env(TRASH_FOLDER)}           ;# Macintosh(?)
	catch {set S(tmp) $env(TMPDIR)}
	catch {set S(tmp) $env(TMP)}
	catch {set S(tmp) $env(TEMP)}
	return $S(tmp)
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set dname "$pfx-$user-$display"
	regsub -all : $dname _ dname
	set fname [file join $dir $dname]
	return $fname
}

proc get_address { fname do_warn} {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			set words [split $line]
			set ipNr [lindex $words 0]
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			return "tcp!$ipNr!$portNr"
		}
		close $fid
	} elseif {$do_warn} {
		warn stderr "$fid"
		return ""
	}
}
		
proc try_connect { address do_warn} {
	set words [split $address !]
	if {[llength $words] == 3 && [string compare [lindex $words 0] tcp] == 0} {
		set host [lindex $words 1]
		set port [lindex $words 2]
		if {[string compare $host 0.0.0.0] == 0} {
			set host 127.0.0.1
		}
		if {![catch {socket $host $port} sock]} {
			return $sock
		} else {
			warn stderr "could not connect tcp!$host!$port: $sock"
		}
	} elseif {$do_warn} {
		warn stderr "could not parse address $address"
	}
	return 0
}

# === end clientsrv.tcl ===== }
# === start process.tcl ===== }
# initially we tried to run 'exec -- tclkit bla -s &' but somehow
# on windows # the '-s' did not arrive with the program that
# we exec-ed
# That's why we tried using '-server' intead of '-s', which does work.

proc start { argv } {
	global tcl_platform starkit topdir argv0 pid
	set pid [pid]
	set tp $tcl_platform(platform)

	package require Tclkit
#	puts "process:$pid: info script: [info script]"
#	puts "process:$pid: starkit::topdir  $starkit::topdir"
#	puts "process:$pid: topdir  $topdir"

#	puts "process:$pid: argv0 $argv0"
#	puts "process:$pid: where we are, as script: driname argv0 : [file dirname $argv0]"
#	puts "process:$pid: where we are, as starpack 1:  dirname of dirname of argv0: [file dirname [file dirname $argv0]]"
#	puts "process:$pid: where we are, as starpack 2: info nameofexecutable:  [info nameofexecutable]"

	set runner ""
	set kit ""
	set cmd {}
	set nohupcmd {}
	set exe_tmp_script {}
	if {[lsearch [info commands] gettclkitpath] >= 0} {
		set exe_tmp_script [gettclkitpath]
	}
	if {[string first [info nameofexecutable] $argv0] == 0} {
#		puts "process:$pid: we are starpack"
		# we are a starpack
		# for the unix server: run [info nameofexecutable] -s
		# for the windows server: run tclkit on server...
		if {[string compare [lindex $exe_tmp_script 0] ""] != 0} {
#			puts "process:$pid: windows: using gettclkitpath"
			set exe [lindex $exe_tmp_script 0]
			set kit [info nameofexecutable]
			set runner [lindex $exe_tmp_script 2]
		} else {
			set exe [info nameofexecutable]
		}
	} else {
		switch -- [file extension $starkit::topdir] {
			.kit {
#				puts "process:$pid: we are ... topdir ~ .kit"
				if {[string compare [lindex $exe_tmp_script 0] ""] != 0} {
					set exe [lindex $exe_tmp_script 0]
				} else {
					set exe [info nameofexecutable]
				}
				set kit $starkit::topdir
			}
			.vfs {
#				puts "process:$pid: we are ... topdir ~ .vfs"
				if {[string compare [lindex $exe_tmp_script 0] ""] != 0} {
					set exe [lindex $exe_tmp_script 0]
				} else {
					set exe [info nameofexecutable]
				}
				set kit [file join $starkit::topdir main.tcl]
			}
			default {
#				puts "process:$pid: we are ... ???"
#				puts "process:$pid: don't know how to start server, just guesing..."
				if {[string compare [lindex $exe_tmp_script 0] ""] != 0} {
					set exe [lindex $exe_tmp_script 0]
				} else {
					set exe [info nameofexecutable]
				}
				set kit $starkit::topdir
				set runner [lindex $exe_tmp_script 2]
			}
		}
	}

	lappend nohupcmd exec --
	lappend cmd exec --
	lappend nohupcmd nohup
	foreach a [list $exe $runner $kit] {
		if {[string compare $a ""] != 0} {
			lappend nohupcmd $a
			lappend cmd $a
		}
	}
	lappend nohupcmd -server $pid &
	lappend cmd -server $pid &



#	set exe_lib [gettclkitpath]
#	set kit [file join [lindex $exe_lib 1] $argv.kit]
#	set exe [lindex $exe_lib 0]
#	set nohupcmd [concat  exec -- nohup $exe $runner $kit -server &]
#	set cmd [concat  exec  -- $exe $runner $kit -server &]
#	puts  "process:$pid: trying to start server: $nohupcmd"
	if {0 && ![catch $nohupcmd msg]} {
#		puts  "process:$pid: trying to start server: result: $msg"
		return
	} elseif {![catch $cmd msg]} {
		puts  "process:$pid: trying to start server: result: $msg"
#		puts  "process:$pid: succeeded starting server: $cmd"
		return
	} else {
		catch {puts stdout "cannot run: $cmd: $msg"}
	}
	return
	switch -- $tp {
	    unix {
		exec sh -c "expnohup $argv &" &
	  } windows {
		exec $argv &
	  }  macintosh {
		warn stderr "don't know how to start $argv (on platform $tp)"
	  } default {
		warn stderr "don't know how to start $argv (unknown platform $tp)"
	  }
	}
}

proc kill {p} {
	global tcl_platform

	msg stderr "cleaning up: kill $p"
	set tp $tcl_platform(platform)
	switch -- $tp {
	    unix {
		catch {exec kill $p}
		msg stderr "cleaning up: kill -9 $p"
		catch {exec kill -9 $p}
	  } windows -
	    macintosh {
		msg stderr "don't know how to kill $p (on platform $tp)"
	  } default {
		msg stderr "don't know how to kill $p (unknown platform $tp)"
	  }
	}
}

proc findprocess {pid p} {
	global tcl_platform

	set found 0
	set tp $tcl_platform(platform)
	warn stdout "findprocess $pid $p  $tp"
	switch -- $tp {
	    unix {
		if {[catch {open "|ps -ef"} pidsfd]} {
			warn stdout "could not run: ps -ef"
			exit 1
		} else {
			while {[gets $pidsfd line] >= 0} {
#				warn stdout "findprocess $pid $p  $tp line $line"
#				debug  "$pid: looking for $p in: $line"
				if {[lindex [split $line] 1] == $p}  {
#					warn stdout "found running server pid: $p"
					set found 1
					break
				}
			}
			# try to catch: child killed: write on pipe with no readers
			catch {close $pidsfd}
		}
	  } windows {
#		warn stdout "findprocess windows"
		if {![catch {open "|tasklist"} pidsfd]} {
#			warn stdout "findprocess windows in tasklist"
			while {[gets $pidsfd line] >= 0} {
				# debug  "$pid: looking for $p in: $line"
				if {[lindex [split $line] 2] == $p}  {
#					warn stdout "tasklist found running server pid: $p"
					set found 1
					break
				}
			}
			close $pidsfd
		} elseif {![catch {open "|tlist"} pidsfd]} {
#			warn stdout "findprocess windows in tlist"
			while {[gets $pidsfd line] >= 0} {
				# debug  "$pid: looking for $p in: $line"
				set nr [lindex [split $line] 0]
				set nr2 [expr int($nr - pow(2,32)]
				if {$nr == $p || $nr2 == $p}  {
#					warn stdout "tlist found running server pid: $p"
					set found 1
					break
				}
			}
			close $pidsfd
		} else {
			# should try tlist on other windows?
			warn stdout "could not run: tasklist"
			exit 1
		}
	  } macintosh {
		warn stdout "don't know how to kill $p (on platform $tp)"
	  } default {
		warn stdout "don't know how to kill $p (unknown platform $tp)"
	  }
	}
	warn stdout "findprocess $pid $p: $found"
	return $found
}

# === end process.tcl ===== }

set debug 0

set G(prog) Bmsc
set G(srvprog) mscviewer
set G(fnamepart)  mscviewer

set prog Bmsc

# we use output to make sure that everything is flushed!
# we do this to avoid deadlocks due to buffers being full...
proc warn { fid s } {
	global PID
	global G
	catch {
		puts $fid "$G(sprog):$PID warning: $s"
		flush $fid
	}
}
proc output {chan msg} {
	catch { puts $chan [string trim $msg] }
	catch { flush $chan }
}

proc fatal {s {code {0}}} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
	cleanup_exit $code
}
proc message {s} {
	global prog
	set s [string trim $s]
	output stdout "$prog: $s"
}
proc warning {s} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
}

proc debug {s} {
	global prog
	global debug
	if {$debug} {
		set s [string trim $s]
		output stderr "$prog: $s"
	}
}

proc cleanup_exit {{code {0}}} {
	global state
	if {[info exists state(stderr)]} {
		eval fconfigure stderr $state(stderr)
	}
	if {[info exists state(stdout)]} {
		eval fconfigure stdout $state(stdout)
	}
	if {[info exists state(stdin)]} {
		eval fconfigure stdin $state(stdin)
	}
	exit $code
}

proc configure_fids {socket} {
	fconfigure $socket -buffering line  -translation {binary binary}
	fconfigure stdin -buffering line -translation {binary binary}
}

proc filevent_fids {socket remoteHost remotePort waitRemoteClose} {
	global  forever
	fileevent $socket readable \
	    	[list getRemote $remoteHost $remotePort $socket]
	fileevent stdin readable \
	 	[list sendRemote $remoteHost $remotePort $socket $waitRemoteClose]
	if {[catch {vwait forever} msg]} {
		fatal "internal error in fileevent_fids: $msg" 1
	}
}


proc getRemote {remoteHost remotePort sock} {
  if {[catch {
  	fileevent $sock readable {}
	set read [gets $sock line]
	debug "getRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
		catch { close stdout }
		cleanup_exit
	} else {
		output stdout $line
		debug "getRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent $sock readable [list getRemote $remoteHost $remotePort $sock]
   } msg]} {
		catch { close stdout }
		cleanup_exit
  }
}
proc sendRemote {remoteHost remotePort sock waitRemoteClose} {
  if {[catch {
	fileevent stdin readable {}
	set read [gets stdin line]
	debug "sendRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
			# set line "A_QUIT"
			# catch {output $sock $line}
			output $sock -eof
			debug "eof stdin"
			if {! $waitRemoteClose} {
				catch {close $sock}
				debug "sendRemote closed tcp!$remoteHost!$remotePort"
				cleanup_exit
			} else {
				return
			}
	} else {
		catch {output $sock $line}
		debug "sendRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent stdin readable [list sendRemote $remoteHost $remotePort $sock $waitRemoteClose]
   } msg]} {
			output $sock -eof
			debug "eof stdin"
			if {! $waitRemoteClose} {
				catch {close $sock}
				debug "sendRemote closed tcp!$remoteHost!$remotePort"
				cleanup_exit
			} else {
				return
			}
   }
}

proc main_loop {sock waitRemoteClose argstring eofstring} {
	configure_fids $sock
	set remote [fconfigure $sock -peername]
	set remoteHost [lindex $remote 0]
	set remotePort [lindex $remote 2]
	puts $sock $argstring
	if {[string compare $eofstring ""] != 0} {
		puts $sock $eofstring
	}
	filevent_fids $sock $remoteHost $remotePort $waitRemoteClose
}

proc usage {} {
	warning "usage: $G(prog) \[-r\] \[-m mcastid\] \[-t title\] \[-k key\] \[-s srvaddr \] \[files.msc\]"
	exit 1
}

proc main {} {
	global argv
	global env
	global pid
	global PID
	global G

	set waitRemoteClose 0

	set reuse ""
	set mcast ""
	set title ""
	set key ""
	set srvaddr ""
	set files {}
	set exit 0
	set errors 0
	set pid [pid]
	set PID $pid

	set argstring ""
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast "-m $env(TORXMCASTID)"
	}

	set args $argv
	
#	set joinedargs [join $argv ")(" ]
#	warning "command lin args: ($joinedargs)"

	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse "-r"
		  } -m {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
		  			set errors 1
				} else {
					set mcast "-m [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
		  			set errors 1
				} else {
					set title "-t {[lindex $args 0]}"
					set args [lrange $args 1 end]
				}
		  } -k {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
		  			set errors 1
				} else {
					set key "-k [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -s {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
					set errors 1
				} else {
					set srvaddr [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -exit {
		  		set exit 1
		  } -* {
		  		warning "unknown flag: $a"
		  			set errors 1
		  } default {
		  		lappend files $a
		  }
		}
	}

#	warning "usage: Bmsc \[-r\] \[-m mcastid\] \[-t title\] \[-k key\] file.dot"
	if {$errors} {
		usage
	}
	set host [info hostname]
	set display [find_display $host]
	set user [find_user]
	set tmp [find_tmp]
	set fname [mk_filename $tmp $G(fnamepart) $user $display]
	# puts stderr "fname=$fname"
	set B(interface,name) $fname
	set B(pidsfile,name) "$fname.pid"
	set B(statusfile,name) "$fname.stat"

	if {[string compare $srvaddr ""] == 0} {
		set srvaddr [get_address $fname 0]
	}
	set sock [try_connect $srvaddr 0]
 
    if {$exit} {
		if {[string compare $sock 0] != 0} {
			puts $sock -exit
			exit
		}
    } else {
	set argstring "-args $reuse $mcast $title $key"
	foreach file $files {
		if {[string compare [file pathtype $file] absolute] != 0} {
			set file [file join [pwd] $file]
		}
		append argstring " $file"
	}
	# puts stderr "sending $argstring"

	if {[llength $files] == 0} {
		set eofstring ""
	} else {
		set eofstring "-eof"
		set waitRemoteClose 1
	}
	
	if {[string compare $sock 0] != 0} {
		main_loop $sock $waitRemoteClose $argstring $eofstring
	} else {
		if {[catch {open $B(statusfile,name) a} fid]} {
			puts  "cannot open (append) $B(statusfile,name): $fid"
			exit 1
 		} else {
			puts $fid "starting $host $pid"
			close $fid
 		}
		if {[catch {open $B(statusfile,name) r} fid]} {
			puts  "cannot open (read) $B(statusfile,name)"
			exit 1
		} elseif {[gets $fid line] < 0} {
			puts  "cannot read $B(statusfile,name)"
			exit 1
		} else {
			close $fid
			#puts  "read $B(statusfile,name): $line"
			set words [split $line]
			set stat [lindex $words 0]
			set h [lindex $words 1]
			set p [lindex $words 2]
			if {[string compare $stat starting] == 0 &&
			    [string compare $host $h] == 0 &&
			    ( [string compare $pid $p] == 0 ||
			      ![findprocess $pid $p])} {
				# puts  "starting  $G(srvprog)"
				start  [list $G(srvprog) $pid]
 			} else {
				puts  "stat $stat (starting) h $h ($host) p $p ($pid)"
			}
			set i 0
			set maxtries 150
			while {$i < $maxtries} {
				if {$i == ($maxtries - 1)} {
					set sock [try_connect [get_address $fname 1] 1]
				} else {
					set sock [try_connect [get_address $fname 0] 0]
				}
				if {[string compare $sock 0] != 0} {
					main_loop $sock $waitRemoteClose $argstring $eofstring
				}
				after 100
				incr i
			}
		}
	}
    }
}

proc start { argv } {
	exec sh -c "expnohup $argv &" &
}

main
exit
