#! @PERL@

$debug = 1;
$debug = 0;
$BUFSIZ = 1024;				# size of chunks we try to read

$| = 1; # do not buffer!

$outpfx=$ARGV[0];
shift(@ARGV);
$errpfx=$ARGV[0];
shift(@ARGV);

$SIG{CHLD} = sub { exit };

$prog=shift(@ARGV);
$progname = $prog;
$progname =~ s,.*/,,;

if ($debug) {
	# open(DEBUG, ">/dev/tty12");
	# $oldfh =  select DEBUG; $| = 1; select $oldfh;
}

start($prog, $progname, @ARGV);
main_loop();

# HACK to work around the problem that we use
#	 --log '|program'
# to start jararacy (visualize test purposes)
# and gnuplot (for coverage of implementation)
#
#if echo "$@" | fgrep '|'
#then
#	((( "$@" ) 3>&- | pfx $out 1>&3) 2>&1 | pfx $err ) 3>&1
#else
#	((( eval "$@" ) 3>&- | pfx $out 1>&3) 2>&1 | pfx $err ) 3>&1
#fi

$SIG{CHLD} = sub { exit };

sub dbgmsg {
	my($d, $s) = @_;
	print STDERR $$." ".$s."\n" if $debug >= $d;
}

sub repr {
	my($s) = @_;

	$s =~ s/\n/\\n/g;
	return $s;
}

sub main_loop {
	my($strReturns) = @_;
	my($ev, $time, $suspension);
	my($buffer,$line,$rest,$nread,$nfound,$stdinbuf,$iutoutbuf,$iuterrbuf);
	my($rin,$win,$ein,$rout,$wout,$eout);
	my(@words,$word);

	# $eofseen = "";

	dbgmsg(1,"STDIN main_loop");
	$timeout = undef;

	while(1) {
		$rin = $win = $ein = "";
		vec($rin, fileno(STDIN), 1 ) = 1 unless $eofseen{STDIN};
		vec($win, fileno(STDOUT), 1) = 1;
		vec($rin, fileno(IUT_OUT), 1) = 1 unless $eofseen{IUT_OUT};
		vec($rin, fileno(IUT_ERR), 1) = 1 unless $eofseen{IUT_ERR};
		$ein = $rin; # | $win;
		# NOTE: can not use nfound: select does not return usable value
		if ($nfound = select($rout=$rin, undef, $eout=$ein, $timeout) < 0) {
			die "select: $!";
		} else {
			# dbgmsg("STDIN main_loop after select");
			dbgmsg(1,"STDIN main_loop after select nfound=$nfound; timeout=$timeout");
		}
		if (vec($eout, fileno(STDIN), 1)) {
			dbgmsg(1,"STDIN err");
				$eofseen{STDIN} = 1;
				close(IUT_IN);
				$timeout = 1;
		}
		if (vec($eout, fileno(IUT_OUT), 1)) {
			dbgmsg(1,"IUT_OUT err");
				$eofseen{IUT_OUT} = 1;
		}
		if (vec($eout, fileno(IUT_ERR), 1)) {
			dbgmsg(1,"IUT_ERR err");
				$eofseen{IUT_ERR} = 1;
		}
		if (vec($rout, fileno(STDIN), 1)) {
			dbgmsg(2,"STDIN ready");
			$buffer = '';
			$stdinbuf .= doread(STDIN);
			dbgmsg(2,"STDIN stdinbuf (".repr($stdinbuf).")");
			$stdinbuf = dowrite(IUT_IN, $stdinbuf, '', $eofseen{STDIN});
			if ($eofseen{STDIN}) {
				close(IUT_IN);
				$timeout = 1;
			}
			dbgmsg(2,"STDIN handled");
		}
		if (vec($rout, fileno(IUT_OUT), 1)) {
			dbgmsg(2,"IUT_OUT ready");
			$iutoutbuf .= doread(IUT_OUT);
			dbgmsg(2,"IUT_OUT iutoutbuf (".repr($iutoutbuf).")");
			$iutoutbuf = dowrite(STDOUT, $iutoutbuf, $outpfx, $eofseen{IUT_OUT});
			dbgmsg(2,"IUT_OUT handled");
		}
		if (vec($rout, fileno(IUT_ERR), 1)) {
			dbgmsg(2,"IUT_ERR ready main_loop");
			$iuterrbuf .= doread(IUT_ERR);
			dbgmsg(2,"IUT_ERR iuterrbuf (".repr($iuterrbuf).")");
			$iuterrbuf = dowrite(STDOUT, $iuterrbuf, $errpfx, $eofseen{IUT_ERR});
			dbgmsg(2,"IUT_ERR handled");
		}
		handle_eof_iut();
		if ($nfound == 0 && defined($timeout)) {
			cleanup;
			exit;
		}
	}
}

sub doread {
	my($fd) = @_;

	my($buffer, $nread) = '';
	if (($nread = sysread($fd, $buffer, $BUFSIZ)) > 0) {
		dbgmsg(2,"$fd read $nread");
	} elsif ($nread == 0) {
		dbgmsg(1,"$fd eof");
		$eofseen{$fd} = 1;
	}
	return $buffer;
}

sub dowrite {
	my($fd, $buf, $pfx, $all) = @_;
	my($line,$rest);

	while ($buf =~ /^([^\n]*\n)/) {
		($line,$rest) = ($1, $');
		# $line =~ s/^[ 	]*//;
		# $line =~ s/[ 	]*$//;
		print $fd $pfx.$line;
		$buf = $rest;
	}
	if ($buf ne "" && $all) {
		$buf = dowrite($fd, $buf."\n", $pfx, 0);
	}
	return $buf;
}

sub handle_eof_iut {
	# AFEB hack: changed && to || to semi-detect eof of torx, also on windows
	if ($eofseen{IUT_OUT} || $eofseen{IUT_ERR}) {
		cleanup;
		exit;
	}
}

sub cleanup {
}


sub start {
	my($cmd, @cmd) = @_;
	my($old, $sh, @shcmd);
	
	dbgmsg(1,"start $cmd : ".join(" ", @cmd));
	$sh = "/bin/sh";
	@shcmd = ($sh, "-c", join(" ", @cmd));
	dbgmsg(2,"shstart $sh : ".join(" ", @shcmd));
	pipe(IUT_OUT,WOFD);    	# Open pipes that will link child and parent
	pipe(IUT_ERR,WEFD);
	pipe(WIFD,IUT_IN);
	if (($cpid = fork()) == 0) {	#in child
		close(IUT_IN);
		close(IUT_OUT);
		close(IUT_ERR);
		close(STDOUT);
		open(STDOUT,">&WOFD");
		$| = 1;						# is this really needed?
		close(STDERR);
		open(STDERR,">&WEFD");
		close(STDIN);
		open(STDIN,"<&WIFD");
		$| = 1;						# is this really needed?
		exec $sh @shcmd;				# do it
		die "exec: $cmd: $!\n";		# or die
	}
	# in parent
	close(WIFD);
	close(WOFD);
	close(WEFD);
    
	# seems we need the following to enable autoflush
	$old = select(IUT_IN);
	$| = 1;
	select($old);
	dbgmsg(1,"started $cmd : ".join(" ", @cmd));
}

# some support routines for the select code.
sub addBits {
    local($fd) = @_;
    vec($fileDescs, $fd, 1) = 1;
    $fds{$fd} = 1;
}
sub delBits {
    local($fd) = @_;
    vec($fileDescs, $fd, 1) = 0;
    delete $fds{$fd};
}
sub allEofBits {
    ! keys %fds;
}
sub printVec {
    local($v) = @_;
    local($i, $result);

    for ($i = (8*length($v)) - 1; $i >= 0; $i--) {
	$result .= (vec($v, $i, 1)) ? "1" : "0";
    }
    $result;
}

